from typing import Any, Dict

from .Action import Action
from .ActionType import ActionType
from ..utils.json_utils import add_attrib
from json import dumps
from azure.functions._durable_functions import _serialize_custom_object
from ..utils.entity_utils import EntityId


class CallEntityAction(Action):
    """Defines the structure of the Call Entity object.

    Provides the information needed by the durable extension to be able to call an activity
    """

    def __init__(self, entity_id: EntityId, operation: str, input_=None):
        self.entity_id: EntityId = entity_id

        # Validating that EntityId exists before trying to parse its instanceId
        if not self.entity_id:
            raise ValueError("entity_id cannot be empty")

        self.instance_id: str = EntityId.get_scheduler_id(entity_id)
        self.operation: str = operation
        self.input_: str = dumps(input_, default=_serialize_custom_object)

    @property
    def action_type(self) -> int:
        """Get the type of action this class represents."""
        return ActionType.CALL_ENTITY

    def to_json(self) -> Dict[str, Any]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Any] = {}
        add_attrib(json_dict, self, "action_type", "actionType")
        add_attrib(json_dict, self, 'instance_id', 'instanceId')
        add_attrib(json_dict, self, 'operation', 'operation')
        add_attrib(json_dict, self, 'input_', 'input')
        return json_dict



from typing import Any, Dict, Union

from .Action import Action
from .ActionType import ActionType
from ..utils.json_utils import add_attrib


class WaitForExternalEventAction(Action):
    """Defines the structure of Wait for External Event object.

    Returns
    -------
    WaitForExternalEventAction
        Returns a WaitForExternalEventAction Class.

    Raises
    ------
    ValueError
        Raises error if external_event_name is not defined.
    """

    def __init__(self, external_event_name: str):
        self.external_event_name: str = external_event_name
        self.reason = "ExternalEvent"

        if not self.external_event_name:
            raise ValueError("external_event_name cannot be empty")

    @property
    def action_type(self) -> int:
        """Get the type of action this class represents."""
        return ActionType.WAIT_FOR_EXTERNAL_EVENT

    def to_json(self) -> Dict[str, Any]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Union[str, int]]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Union[str, int]] = {}

        add_attrib(json_dict, self, 'action_type', 'actionType')
        add_attrib(json_dict, self, 'external_event_name', 'externalEventName')
        add_attrib(json_dict, self, 'reason', 'reason')
        return json_dict

    def __eq__(self, other):
        """Override the default __eq__ method.

        Returns
        -------
        Bool
            Returns True if two class instances has same values at all properties,
            and returns False otherwise.
        """
        if not isinstance(other, WaitForExternalEventAction):
            return False
        else:
            return self.action_type == other.action_type \
                and self.external_event_name == other.external_event_name \
                and self.reason == other.reason



"""Defines the models for the different forms of Activities that can be scheduled."""
from .Action import Action
from .ActionType import ActionType
from .CallActivityAction import CallActivityAction
from .CallActivityWithRetryAction import CallActivityWithRetryAction
from .CallSubOrchestratorAction import CallSubOrchestratorAction
from .WaitForExternalEventAction import WaitForExternalEventAction
from .CallHttpAction import CallHttpAction
from .CreateTimerAction import CreateTimerAction
from .WhenAllAction import WhenAllAction
from .WhenAnyAction import WhenAnyAction

__all__ = [
    'Action',
    'ActionType',
    'CallActivityAction',
    'CallActivityWithRetryAction',
    'CallSubOrchestratorAction',
    'CallHttpAction',
    'WaitForExternalEventAction',
    'CreateTimerAction',
    'WhenAnyAction',
    'WhenAllAction'
]



from typing import Optional, Dict, Any
from azure.functions._durable_functions import _serialize_custom_object
import json


class OperationResult:
    """OperationResult.

    The result of an Entity operation.
    """

    def __init__(self,
                 is_error: bool,
                 duration: int,
                 result: Optional[str] = None):
        """Instantiate an OperationResult.

        Parameters
        ----------
        is_error: bool
            Whether or not the operation resulted in an exception.
        duration: int
            How long the operation took, in milliseconds.
        result: Optional[str]
            The operation result. Defaults to None.
        """
        self._is_error: bool = is_error
        self._duration: int = duration
        self._result: Optional[str] = result

    @property
    def is_error(self) -> bool:
        """Determine if the operation resulted in an error.

        Returns
        -------
        bool
            True if the operation resulted in error. Otherwise False.
        """
        return self._is_error

    @property
    def duration(self) -> int:
        """Get the duration of this operation.

        Returns
        -------
        int:
            The duration of this operation, in milliseconds
        """
        return self._duration

    @property
    def result(self) -> Any:
        """Get the operation's result.

        Returns
        -------
        Any
            The operation's result
        """
        return self._result

    def to_json(self) -> Dict[str, Any]:
        """Represent OperationResult as a JSON-serializable Dict.

        Returns
        -------
        Dict[str, Any]
            A JSON-serializable Dict of the OperationResult
        """
        to_json: Dict[str, Any] = {}
        to_json["isError"] = self.is_error
        to_json["duration"] = self.duration
        to_json["result"] = json.dumps(self.result, default=_serialize_custom_object)
        return to_json



# This function an HTTP starter function for Durable Functions.
# Before running this sample, please:
# - create a Durable orchestration function
# - create a Durable activity function (default name is "Hello")
# - add azure-functions-durable to requirements.txt
# - run pip install -r requirements.txt
 
import logging
import json
import azure.functions as func
import azure.durable_functions as df

async def main(req: func.HttpRequest, starter: str) -> func.HttpResponse:
    client = df.DurableOrchestrationClient(starter)

    payload: str = json.loads(req.get_body().decode())
    instance_id = await client.start_new(req.route_params["functionName"], client_input=payload)

    logging.info(f"Started orchestration with ID = '{instance_id}'.")

    return client.create_check_status_response(req, instance_id)


import os
from os.path import dirname
from typing import List

def main(rootDirectory: str) -> List[str]:

    all_file_paths = []
    # We walk the file system
    for path, _, files in os.walk(rootDirectory):
        # We copy the code for activities and orchestrators
        if "E2_" in path:
            # For each file, we add their full-path to the list
            for name in files:
                if name == "__init__.py" or name == "function.json":
                    file_path = os.path.join(path, name)
                    all_file_paths.append(file_path)
    
    return all_file_paths



from typing import Optional, Any, Dict, Tuple, List, Callable
from azure.functions._durable_functions import _deserialize_custom_object
import json


class DurableEntityContext:
    """Context of the durable entity context.

    Describes the API used to specify durable entity user code.
    """

    def __init__(self,
                 name: str,
                 key: str,
                 exists: bool,
                 state: Any):
        """Context of the durable entity context.

        Describes the API used to specify durable entity user code.

        Parameters
        ----------
        name: str
            The name of the Durable Entity
        key: str
            The key of the Durable Entity
        exists: bool
            Flag to determine if the entity exists
        state: Any
            The internal state of the Durable Entity
        """
        self._entity_name: str = name
        self._entity_key: str = key

        self._exists: bool = exists
        self._is_newly_constructed: bool = False

        self._state: Any = state
        self._input: Any = None
        self._operation: Optional[str] = None
        self._result: Any = None

    @property
    def entity_name(self) -> str:
        """Get the name of the Entity.

        Returns
        -------
        str
            The name of the entity
        """
        return self._entity_name

    @property
    def entity_key(self) -> str:
        """Get the Entity key.

        Returns
        -------
        str
            The entity key
        """
        return self._entity_key

    @property
    def operation_name(self) -> Optional[str]:
        """Get the current operation name.

        Returns
        -------
        Optional[str]
            The current operation name
        """
        if self._operation is None:
            raise Exception("Entity operation is unassigned")
        return self._operation

    @property
    def is_newly_constructed(self) -> bool:
        """Determine if the Entity was newly constructed.

        Returns
        -------
        bool
            True if the Entity was newly constructed. False otherwise.
        """
        # This is not updated at the moment, as its semantics are unclear
        return self._is_newly_constructed

    @classmethod
    def from_json(cls, json_str: str) -> Tuple['DurableEntityContext', List[Dict[str, Any]]]:
        """Instantiate a DurableEntityContext from a JSON-formatted string.

        Parameters
        ----------
        json_string: str
            A JSON-formatted string, returned by the durable-extension,
            which represents the entity context

        Returns
        -------
        DurableEntityContext
            The DurableEntityContext originated from the input string
        """
        json_dict = json.loads(json_str)
        json_dict["name"] = json_dict["self"]["name"]
        json_dict["key"] = json_dict["self"]["key"]
        json_dict.pop("self")

        serialized_state = json_dict["state"]
        if serialized_state is not None:
            json_dict["state"] = from_json_util(serialized_state)

        batch = json_dict.pop("batch")
        return cls(**json_dict), batch

    def set_state(self, state: Any) -> None:
        """Set the state of the entity.

        Parameter
        ---------
        state: Any
            The new state of the entity
        """
        self._exists = True

        # should only serialize the state at the end of the batch
        self._state = state

    def get_state(self, initializer: Optional[Callable[[], Any]] = None) -> Any:
        """Get the current state of this entity.

        Parameters
        ----------
        initializer: Optional[Callable[[], Any]]
            A 0-argument function to provide an initial state. Defaults to None.

        Returns
        -------
        Any
            The current state of the entity
        """
        state = self._state
        if state is not None:
            return state
        elif initializer:
            if not callable(initializer):
                raise Exception("initializer argument needs to be a callable function")
            state = initializer()
        return state

    def get_input(self) -> Any:
        """Get the input for this operation.

        Returns
        -------
        Any
            The input for the current operation
        """
        input_ = None
        req_input = self._input
        req_input = json.loads(req_input)
        input_ = None if req_input is None else from_json_util(req_input)
        return input_

    def set_result(self, result: Any) -> None:
        """Set the result (return value) of the entity.

        Paramaters
        ----------
        result: Any
            The result / return value for the entity
        """
        self._exists = True
        self._result = result

    def destruct_on_exit(self) -> None:
        """Delete this entity after the operation completes."""
        self._exists = False
        self._state = None


def from_json_util(json_str: str) -> Any:
    """Load an arbitrary datatype from its JSON representation.

    The Out-of-proc SDK has a special JSON encoding strategy
    to enable arbitrary datatypes to be serialized. This utility
    loads a JSON with the assumption that it follows that encoding
    method.

    Parameters
    ----------
    json_str: str
        A JSON-formatted string, from durable-extension

    Returns
    -------
    Any:
        The original datatype that was serialized
    """
    return json.loads(json_str, object_hook=_deserialize_custom_object)



from ..utils.entity_utils import EntityId


class Signal:
    """An EntitySignal.

    Describes a signal call to a Durable Entity.
    """

    def __init__(self,
                 target: EntityId,
                 name: str,
                 input_: str):
        """Instantiate an EntitySignal.

        Instantiate a signal call to a Durable Entity.

        Parameters
        ----------
        target: EntityId
            The target of signal
        name: str
            The name of the signal
        input_: str
            The signal's input
        """
        self._target = target
        self._name = name
        self._input = input_

    @property
    def target(self) -> EntityId:
        """Get the Signal's target entity.

        Returns
        -------
        EntityId
            EntityId of the target
        """
        return self._target

    @property
    def name(self) -> str:
        """Get the Signal's name.

        Returns
        -------
        str
            The Signal's name
        """
        return self._name

    @property
    def input(self) -> str:
        """Get the Signal's input.

        Returns
        -------
        str
            The Signal's input
        """
        return self._input



"""Unit tests for the durable functions library"""
import os
import sys
import unittest


def suite():
    """

    :return: configuration for the suite of tests
    """
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover(
        os.path.dirname(__file__), pattern='test_*.py')
    return test_suite


if __name__ == '__main__':
    runner = unittest.runner.TextTestRunner()
    result = runner.run(suite())
    sys.exit(not result.wasSuccessful())



class FunctionContext:
    """Object to hold any additional function level attributes not used by Durable."""

    def __init__(self, **kwargs):
        if kwargs is not None:
            for key, value in kwargs.items():
                self.__setattr__(key, value)



import json
import random

random.seed(10)

def main(phoneNumber, message):
  code = random.randint(0, 10000)
  payload = {
    "body": f"Your verification code is {code}",
    "to": phoneNumber
  }

  message.set(json.dumps(payload))
  code_str = str(code)
  return code_str



from tests.conftest import TASK_HUB_NAME, replace_stand_in_bits


def test_extracts_task_hub_name(binding_info):
    assert TASK_HUB_NAME == binding_info.task_hub_name


def test_extracts_create_new_instance_post_uri(binding_info):
    expected_url = replace_stand_in_bits(
        "BASE_URL/orchestrators/{functionName}[/{instanceId}]?code=AUTH_CODE")
    assert \
        expected_url == binding_info.creation_urls["createNewInstancePostUri"]


def test_extracts_create_and_wait_on_new_instance_post_uri(binding_info):
    expected_url = replace_stand_in_bits(
        "BASE_URL/orchestrators/{functionName}[/{instanceId}]?timeout={"
        "timeoutInSeconds}&pollingInterval={intervalInSeconds}&code=AUTH_CODE")
    assert expected_url == binding_info.creation_urls[
        "createAndWaitOnNewInstancePostUri"]


def test_extracts_status_query_get_uri(binding_info):
    expected_url = replace_stand_in_bits(
        "BASE_URL/instances/INSTANCEID?taskHub=TASK_HUB_NAME&connection"
        "=Storage&code=AUTH_CODE")
    assert expected_url == binding_info.management_urls["statusQueryGetUri"]


def test_extracts_send_event_post_uri(binding_info):
    expected_url = replace_stand_in_bits(
        "BASE_URL/instances/INSTANCEID/raiseEvent/{"
        "eventName}?taskHub=TASK_HUB_NAME&connection=Storage&code=AUTH_CODE")
    assert expected_url == binding_info.management_urls["sendEventPostUri"]


def test_extracts_terminate_post_uri(binding_info):
    expected_url = replace_stand_in_bits(
        "BASE_URL/instances/INSTANCEID/terminate?reason={"
        "text}&taskHub=TASK_HUB_NAME&connection=Storage&code=AUTH_CODE")
    assert expected_url == binding_info.management_urls["terminatePostUri"]


def test_extracts_rewind_post_uri(binding_info):
    expected_url = replace_stand_in_bits(
        "BASE_URL/instances/INSTANCEID/rewind?reason={"
        "text}&taskHub=TASK_HUB_NAME&connection=Storage&code=AUTH_CODE")
    assert expected_url == binding_info.management_urls["rewindPostUri"]


def test_extracts_purge_history_delete_uri(binding_info):
    expected_url = replace_stand_in_bits(
        "BASE_URL/instances/INSTANCEID?taskHub=TASK_HUB_NAME&connection"
        "=Storage&code=AUTH_CODE")
    assert expected_url == binding_info.management_urls[
        "purgeHistoryDeleteUri"]



from azure.durable_functions.models.ReplaySchema import ReplaySchema
from typing import List

from azure.durable_functions.models.actions.Action import Action
from azure.durable_functions.models.actions.CallActivityAction \
    import CallActivityAction
from azure.durable_functions.models.OrchestratorState import OrchestratorState


def test_empty_state_to_json_string():
    actions: List[List[Action]] = []
    state = OrchestratorState(is_done=False, actions=actions, output=None, replay_schema=ReplaySchema.V1.value)
    result = state.to_json_string()
    expected_result = '{"isDone": false, "actions": []}'
    assert expected_result == result


def test_single_action_state_to_json_string():
    actions: List[List[Action]] = []
    action: Action = CallActivityAction(
        function_name="MyFunction", input_="AwesomeInput")
    actions.append([action])
    state = OrchestratorState(is_done=False, actions=actions, output=None, replay_schema=ReplaySchema.V1.value)
    result = state.to_json_string()
    expected_result = ('{"isDone": false, "actions": [[{"actionType": 0, '
                       '"functionName": "MyFunction", "input": '
                       '"\\"AwesomeInput\\""}]]}')
    assert expected_result == result



import json
from datetime import datetime
from typing import List, Any, Optional, Dict, Union
from time import time
from asyncio import sleep
from urllib.parse import urlparse, quote

import azure.functions as func

from .PurgeHistoryResult import PurgeHistoryResult
from .DurableOrchestrationStatus import DurableOrchestrationStatus
from .EntityStateResponse import EntityStateResponse
from .RpcManagementOptions import RpcManagementOptions
from .OrchestrationRuntimeStatus import OrchestrationRuntimeStatus
from ..models.DurableOrchestrationBindings import DurableOrchestrationBindings
from .utils.http_utils import get_async_request, post_async_request, delete_async_request
from .utils.entity_utils import EntityId
from azure.functions._durable_functions import _serialize_custom_object


class DurableOrchestrationClient:
    """Durable Orchestration Client.

    Client for starting, querying, terminating and raising events to
    orchestration instances.
    """

    def __init__(self, context: str):
        self.task_hub_name: str
        self._uniqueWebHookOrigins: List[str]
        self._event_name_placeholder: str = "{eventName}"
        self._function_name_placeholder: str = "{functionName}"
        self._instance_id_placeholder: str = "[/{instanceId}]"
        self._reason_placeholder: str = "{text}"
        self._created_time_from_query_key: str = "createdTimeFrom"
        self._created_time_to_query_key: str = "createdTimeTo"
        self._runtime_status_query_key: str = "runtimeStatus"
        self._show_history_query_key: str = "showHistory"
        self._show_history_output_query_key: str = "showHistoryOutput"
        self._show_input_query_key: str = "showInput"
        self._orchestration_bindings: DurableOrchestrationBindings = \
            DurableOrchestrationBindings.from_json(context)
        self._post_async_request = post_async_request
        self._get_async_request = get_async_request
        self._delete_async_request = delete_async_request

    async def start_new(self,
                        orchestration_function_name: str,
                        instance_id: Optional[str] = None,
                        client_input: Optional[Any] = None) -> str:
        """Start a new instance of the specified orchestrator function.

        If an orchestration instance with the specified ID already exists, the
        existing instance will be silently replaced by this new instance.

        Parameters
        ----------
        orchestration_function_name : str
            The name of the orchestrator function to start.
        instance_id : Optional[str]
            The ID to use for the new orchestration instance. If no instance id is specified,
            the Durable Functions extension will generate a random GUID (recommended).
        client_input : Optional[Any]
            JSON-serializable input value for the orchestrator function.

        Returns
        -------
        str
            The ID of the new orchestration instance if successful, None if not.
        """
        request_url = self._get_start_new_url(
            instance_id=instance_id, orchestration_function_name=orchestration_function_name)

        response: List[Any] = await self._post_async_request(
            request_url, self._get_json_input(client_input))

        status_code: int = response[0]
        if status_code <= 202 and response[1]:
            return response[1]["id"]
        elif status_code == 400:
            # Orchestrator not found, report clean exception
            exception_data: Dict[str, str] = response[1]
            exception_message = exception_data["ExceptionMessage"]
            raise Exception(exception_message)
        else:
            # Catch all: simply surfacing the durable-extension exception
            # we surface the stack trace too, since this may be a more involed exception
            ex_message: Any = response[1]
            raise Exception(ex_message)

    def create_check_status_response(
            self, request: func.HttpRequest, instance_id: str) -> func.HttpResponse:
        """Create a HttpResponse that contains useful information for \
        checking the status of the specified instance.

        Parameters
        ----------
        request : HttpRequest
            The HTTP request that triggered the current orchestration instance.
        instance_id : str
            The ID of the orchestration instance to check.

        Returns
        -------
        HttpResponse
           An HTTP 202 response with a Location header
           and a payload containing instance management URLs
        """
        http_management_payload = self.get_client_response_links(request, instance_id)
        response_args = {
            "status_code": 202,
            "body": json.dumps(http_management_payload),
            "headers": {
                "Content-Type": "application/json",
                "Location": http_management_payload["statusQueryGetUri"],
                "Retry-After": "10",
            },
        }
        return func.HttpResponse(**response_args)

    def create_http_management_payload(self, instance_id: str) -> Dict[str, str]:
        """Create a dictionary of orchestrator management urls.

        Parameters
        ----------
        instance_id : str
            The ID of the orchestration instance to check.

        Returns
        -------
        Dict[str, str]
            a dictionary object of orchestrator instance management urls
        """
        return self.get_client_response_links(None, instance_id)

    async def read_entity_state(
        self,
        entityId: EntityId,
        task_hub_name: Optional[str] = None,
        connection_name: Optional[str] = None,
    ) -> EntityStateResponse:
        """Read the state of the entity.

        Parameters
        ----------
        entityId : EntityId
            The EntityId of the targeted entity.
        task_hub_name : Optional[str]
            The task hub name of the target entity.
        connection_name : Optional[str]
            The name of the connection string associated with [task_hub_name].

        Raises
        ------
        Exception:
            When an unexpected status code is returned

        Returns
        -------
        EntityStateResponse
            container object representing the state of the entity
        """
        options = RpcManagementOptions(
            connection_name=connection_name,
            task_hub_name=task_hub_name,
            entity_Id=entityId,
        )

        request_url = options.to_url(self._orchestration_bindings.rpc_base_url)
        response = await self._get_async_request(request_url)

        switch_statement = {
            200: lambda: EntityStateResponse(True, response[1]),
            404: lambda: EntityStateResponse(False),
        }

        result = switch_statement.get(response[0])

        if not result:
            raise Exception(
                f"The operation failed with an unexpected status code {response[0]}"
            )

        return result()

    def get_client_response_links(
            self,
            request: Optional[func.HttpRequest], instance_id: str) -> Dict[str, str]:
        """Create a dictionary of orchestrator management urls.

        Parameters
        ----------
        request : Optional[HttpRequest]
            The HTTP request that triggered the current orchestration instance.
        instance_id : str
            The ID of the orchestration instance to check.

        Returns
        -------
        Dict[str, str]
            a dictionary object of orchestrator instance management urls
        """
        payload = self._orchestration_bindings.management_urls.copy()

        for key, _ in payload.items():
            if not (request is None) and request.url:
                payload[key] = self._replace_url_origin(request.url, payload[key])
            payload[key] = payload[key].replace(
                self._orchestration_bindings.management_urls["id"], instance_id)

        return payload

    async def raise_event(
            self, instance_id: str, event_name: str, event_data: Any = None,
            task_hub_name: str = None, connection_name: str = None) -> None:
        """Send an event notification message to a waiting orchestration instance.

        In order to handle the event, the target orchestration instance must be
        waiting for an event named `eventName` using waitForExternalEvent API.

        Parameters
        ----------
        instance_id : str
            The ID of the orchestration instance that will handle the event.
        event_name : str
            The name of the event.
        event_data : Any, optional
            The JSON-serializable data associated with the event.
        task_hub_name : str, optional
            The TaskHubName of the orchestration that will handle the event.
        connection_name : str, optional
            The name of the connection string associated with `taskHubName.`

        Raises
        ------
        ValueError
            event name must be a valid string.
        Exception
            Raises an exception if the status code is 404 or 400 when raising the event.
        """
        if event_name == "":
            raise ValueError("event_name must be a non-empty string.")

        request_url = self._get_raise_event_url(
            instance_id, event_name, task_hub_name, connection_name)

        response = await self._post_async_request(request_url, json.dumps(event_data))

        switch_statement = {
            202: lambda: None,
            410: lambda: f"Instance with ID {instance_id} is gone: either completed or failed",
            404: lambda: f"No instance with ID {instance_id} found.",
            400: lambda: "Only application/json request content is supported"
        }
        has_error_message = switch_statement.get(
            response[0], lambda: f"Webhook returned unrecognized status code {response[0]}")
        error_message = has_error_message()
        if error_message:
            raise Exception(error_message)

    async def get_status(self, instance_id: str, show_history: bool = False,
                         show_history_output: bool = False,
                         show_input: bool = False) -> DurableOrchestrationStatus:
        """Get the status of the specified orchestration instance.

        Parameters
        ----------
        instance_id : str
            The ID of the orchestration instance to query.
        show_history: bool
            Boolean marker for including execution history in the response.
        show_history_output: bool
            Boolean marker for including output in the execution history response.
        show_input: bool
            Boolean marker for including the input in the response.

        Returns
        -------
        DurableOrchestrationStatus
            The status of the requested orchestration instance
        """
        options = RpcManagementOptions(instance_id=instance_id, show_history=show_history,
                                       show_history_output=show_history_output,
                                       show_input=show_input)
        request_url = options.to_url(self._orchestration_bindings.rpc_base_url)
        response = await self._get_async_request(request_url)
        switch_statement = {
            200: lambda: None,  # instance completed
            202: lambda: None,  # instance in progress
            400: lambda: None,  # instance failed or terminated
            404: lambda: None,  # instance not found or pending
            500: lambda: None  # instance failed with unhandled exception
        }

        has_error_message = switch_statement.get(
            response[0],
            lambda: f"The operation failed with an unexpected status code {response[0]}")
        error_message = has_error_message()
        if error_message:
            raise Exception(error_message)
        else:
            return DurableOrchestrationStatus.from_json(response[1])

    async def get_status_all(self) -> List[DurableOrchestrationStatus]:
        """Get the status of all orchestration instances.

        Returns
        -------
        DurableOrchestrationStatus
            The status of the requested orchestration instances
        """
        options = RpcManagementOptions()
        request_url = options.to_url(self._orchestration_bindings.rpc_base_url)
        response = await self._get_async_request(request_url)
        switch_statement = {
            200: lambda: None,  # instance completed
        }

        has_error_message = switch_statement.get(
            response[0],
            lambda: f"The operation failed with an unexpected status code {response[0]}")
        error_message = has_error_message()
        if error_message:
            raise Exception(error_message)
        else:
            statuses: List[Any] = response[1]
            return [DurableOrchestrationStatus.from_json(o) for o in statuses]

    async def get_status_by(self, created_time_from: datetime = None,
                            created_time_to: datetime = None,
                            runtime_status: List[OrchestrationRuntimeStatus] = None) \
            -> List[DurableOrchestrationStatus]:
        """Get the status of all orchestration instances that match the specified conditions.

        Parameters
        ----------
        created_time_from : datetime
            Return orchestration instances which were created after this Date.
        created_time_to: datetime
            Return orchestration instances which were created before this Date.
        runtime_status: List[OrchestrationRuntimeStatus]
            Return orchestration instances which match any of the runtimeStatus values
            in this list.

        Returns
        -------
        DurableOrchestrationStatus
            The status of the requested orchestration instances
        """
        # TODO: do we really want folks to us this without specifying all the args?
        options = RpcManagementOptions(created_time_from=created_time_from,
                                       created_time_to=created_time_to,
                                       runtime_status=runtime_status)
        request_url = options.to_url(self._orchestration_bindings.rpc_base_url)
        response = await self._get_async_request(request_url)
        switch_statement = {
            200: lambda: None,  # instance completed
        }

        has_error_message = switch_statement.get(
            response[0],
            lambda: f"The operation failed with an unexpected status code {response[0]}")
        error_message = has_error_message()
        if error_message:
            raise Exception(error_message)
        else:
            return [DurableOrchestrationStatus.from_json(o) for o in response[1]]

    async def purge_instance_history(self, instance_id: str) -> PurgeHistoryResult:
        """Delete the history of the specified orchestration instance.

        Parameters
        ----------
        instance_id : str
            The ID of the orchestration instance to delete.

        Returns
        -------
        PurgeHistoryResult
            The results of the request to delete the orchestration instance
        """
        request_url = f"{self._orchestration_bindings.rpc_base_url}instances/{instance_id}"
        response = await self._delete_async_request(request_url)
        return self._parse_purge_instance_history_response(response)

    async def purge_instance_history_by(
            self, created_time_from: Optional[datetime] = None,
            created_time_to: Optional[datetime] = None,
            runtime_status: Optional[List[OrchestrationRuntimeStatus]] = None) \
            -> PurgeHistoryResult:
        """Delete the history of all orchestration instances that match the specified conditions.

        Parameters
        ----------
        created_time_from : Optional[datetime]
            Delete orchestration history which were created after this Date.
        created_time_to: Optional[datetime]
            Delete orchestration history which were created before this Date.
        runtime_status: Optional[List[OrchestrationRuntimeStatus]]
            Delete orchestration instances which match any of the runtimeStatus values
            in this list.

        Returns
        -------
        PurgeHistoryResult
            The results of the request to purge history
        """
        options = RpcManagementOptions(created_time_from=created_time_from,
                                       created_time_to=created_time_to,
                                       runtime_status=runtime_status)
        request_url = options.to_url(self._orchestration_bindings.rpc_base_url)
        response = await self._delete_async_request(request_url)
        return self._parse_purge_instance_history_response(response)

    async def terminate(self, instance_id: str, reason: str) -> None:
        """Terminate the specified orchestration instance.

        Parameters
        ----------
        instance_id : str
            The ID of the orchestration instance to query.
        reason: str
            The reason for terminating the instance.

        Raises
        ------
        Exception:
            When the terminate call failed with an unexpected status code

        Returns
        -------
        None
        """
        request_url = f"{self._orchestration_bindings.rpc_base_url}instances/{instance_id}/" \
                      f"terminate?reason={quote(reason)}"
        response = await self._post_async_request(request_url, None)
        switch_statement = {
            202: lambda: None,  # instance in progress
            410: lambda: None,  # instance failed or terminated
            404: lambda: f"No instance with ID '{instance_id}' found.",
        }

        has_error_message = switch_statement.get(
            response[0],
            lambda: f"The operation failed with an unexpected status code {response[0]}")
        error_message = has_error_message()
        if error_message:
            raise Exception(error_message)

    async def wait_for_completion_or_create_check_status_response(
            self, request, instance_id: str, timeout_in_milliseconds: int = 10000,
            retry_interval_in_milliseconds: int = 1000) -> func.HttpResponse:
        """Create an HTTP response.

        The response either contains a payload of management URLs for a non-completed instance or
        contains the payload containing the output of the completed orchestration.

        If the orchestration does not complete within the specified timeout, then the HTTP response
        will be identical to that of [[createCheckStatusResponse]].

        Parameters
        ----------
        request
            The HTTP request that triggered the current function.
        instance_id:
            The unique ID of the instance to check.
        timeout_in_milliseconds:
            Total allowed timeout for output from the durable function.
            The default value is 10 seconds.
        retry_interval_in_milliseconds:
            The timeout between checks for output from the durable function.
            The default value is 1 second.
        """
        if retry_interval_in_milliseconds > timeout_in_milliseconds:
            raise Exception(f'Total timeout {timeout_in_milliseconds} (ms) should be bigger than '
                            f'retry timeout {retry_interval_in_milliseconds} (ms)')

        checking = True
        start_time = time()

        while checking:
            status = await self.get_status(instance_id)

            if status:
                switch_statement = {
                    OrchestrationRuntimeStatus.Completed:
                        lambda: self._create_http_response(200, status.output),
                    OrchestrationRuntimeStatus.Canceled:
                        lambda: self._create_http_response(200, status.to_json()),
                    OrchestrationRuntimeStatus.Terminated:
                        lambda: self._create_http_response(200, status.to_json()),
                    OrchestrationRuntimeStatus.Failed:
                        lambda: self._create_http_response(500, status.to_json()),
                    None:
                        None
                }

                result = switch_statement.get(status.runtime_status)
                if result:
                    return result()

            elapsed = time() - start_time
            elapsed_in_milliseconds = elapsed * 1000
            if elapsed_in_milliseconds < timeout_in_milliseconds:
                remaining_time = timeout_in_milliseconds - elapsed_in_milliseconds
                sleep_time = retry_interval_in_milliseconds \
                    if remaining_time > retry_interval_in_milliseconds else remaining_time
                sleep_time /= 1000
                await sleep(sleep_time)
            else:
                return self.create_check_status_response(request, instance_id)
        return self.create_check_status_response(request, instance_id)

    async def signal_entity(self, entityId: EntityId, operation_name: str,
                            operation_input: Optional[Any] = None,
                            task_hub_name: Optional[str] = None,
                            connection_name: Optional[str] = None) -> None:
        """Signals an entity to perform an operation.

        Parameters
        ----------
        entityId : EntityId
            The EntityId of the targeted entity to perform operation.
        operation_name: str
            The name of the operation.
        operation_input: Optional[Any]
            The content for the operation.
        task_hub_name: Optional[str]
            The task hub name of the target entity.
        connection_name: Optional[str]
            The name of the connection string associated with [task_hub_name].

        Raises
        ------
        Exception:
            When the signal entity call failed with an unexpected status code

        Returns
        -------
        None
        """
        options = RpcManagementOptions(operation_name=operation_name,
                                       connection_name=connection_name,
                                       task_hub_name=task_hub_name,
                                       entity_Id=entityId)

        request_url = options.to_url(self._orchestration_bindings.rpc_base_url)
        response = await self._post_async_request(
            request_url,
            json.dumps(operation_input) if operation_input else None)

        switch_statement = {
            202: lambda: None  # signal accepted
        }

        has_error_message = switch_statement.get(
            response[0],
            lambda: f"The operation failed with an unexpected status code {response[0]}")

        error_message = has_error_message()

        if error_message:
            raise Exception(error_message)

    @staticmethod
    def _create_http_response(
            status_code: int, body: Union[str, Any]) -> func.HttpResponse:
        body_as_json = body if isinstance(body, str) else json.dumps(body)
        response_args = {
            "status_code": status_code,
            "body": body_as_json,
            "mimetype": "application/json",
            "headers": {
                "Content-Type": "application/json",
            }
        }
        return func.HttpResponse(**response_args)

    @staticmethod
    def _get_json_input(client_input: object) -> Optional[str]:
        """Serialize the orchestrator input.

        Parameters
        ----------
        client_input: object
            The client's input, which we need to serialize

        Returns
        -------
        Optional[str]
            If `client_input` is not None, return a string representing
            the JSON-serialization of `client_input`. Otherwise, returns
            None

        Exceptions
        ----------
        TypeError
            If the JSON serialization failed, see `serialize_custom_object`
        """
        if client_input is not None:
            return json.dumps(client_input, default=_serialize_custom_object)
        return None

    @staticmethod
    def _replace_url_origin(request_url: str, value_url: str) -> str:
        request_parsed_url = urlparse(request_url)
        value_parsed_url = urlparse(value_url)
        request_url_origin = '{url.scheme}://{url.netloc}/'.format(url=request_parsed_url)
        value_url_origin = '{url.scheme}://{url.netloc}/'.format(url=value_parsed_url)
        value_url = value_url.replace(value_url_origin, request_url_origin)
        return value_url

    @staticmethod
    def _parse_purge_instance_history_response(
            response: List[Any]) -> PurgeHistoryResult:
        switch_statement = {
            200: lambda: PurgeHistoryResult.from_json(response[1]),  # instance completed
            404: lambda: PurgeHistoryResult(instancesDeleted=0),  # instance not found
        }

        switch_result = switch_statement.get(
            response[0],
            lambda: f"The operation failed with an unexpected status code {response[0]}")
        result = switch_result()
        if isinstance(result, PurgeHistoryResult):
            return result
        else:
            raise Exception(result)

    def _get_start_new_url(
            self, instance_id: Optional[str], orchestration_function_name: str) -> str:
        instance_path = f'/{instance_id}' if instance_id is not None else ''
        request_url = f'{self._orchestration_bindings.rpc_base_url}orchestrators/' \
                      f'{orchestration_function_name}{instance_path}'
        return request_url

    def _get_raise_event_url(
            self, instance_id: str, event_name: str,
            task_hub_name: Optional[str], connection_name: Optional[str]) -> str:
        request_url = f'{self._orchestration_bindings.rpc_base_url}' \
                      f'instances/{instance_id}/raiseEvent/{event_name}'

        query: List[str] = []
        if task_hub_name:
            query.append(f'taskHub={task_hub_name}')

        if connection_name:
            query.append(f'connection={connection_name}')

        if len(query) > 0:
            request_url += "?" + "&".join(query)

        return request_url

    async def rewind(self,
                     instance_id: str,
                     reason: str,
                     task_hub_name: Optional[str] = None,
                     connection_name: Optional[str] = None):
        """Return / "rewind" a failed orchestration instance to a prior "healthy" state.

        Parameters
        ----------
        instance_id: str
            The ID of the orchestration instance to rewind.
        reason: str
            The reason for rewinding the orchestration instance.
        task_hub_name: Optional[str]
            The TaskHub of the orchestration to rewind
        connection_name: Optional[str]
            Name of the application setting containing the storage
            connection string to use.

        Raises
        ------
        Exception:
            In case of a failure, it reports the reason for the exception
        """
        request_url: str = ""
        if self._orchestration_bindings.rpc_base_url:
            path = f"instances/{instance_id}/rewind?reason={reason}"
            query: List[str] = []
            if not (task_hub_name is None):
                query.append(f"taskHub={task_hub_name}")
            if not (connection_name is None):
                query.append(f"connection={connection_name}")
            if len(query) > 0:
                path += "&" + "&".join(query)

            request_url = f"{self._orchestration_bindings.rpc_base_url}" + path
        else:
            raise Exception("The Python SDK only supports RPC endpoints."
                            + "Please remove the `localRpcEnabled` setting from host.json")

        response = await self._post_async_request(request_url, None)
        status: int = response[0]
        ex_msg: str = ""
        if status == 200 or status == 202:
            return
        elif status == 404:
            ex_msg = f"No instance with ID {instance_id} found."
            raise Exception(ex_msg)
        elif status == 410:
            ex_msg = "The rewind operation is only supported on failed orchestration instances."
            raise Exception(ex_msg)
        elif isinstance(response[1], str):
            ex_msg = response[1]
            raise Exception(ex_msg)
        else:
            ex_msg = "Received unexpected payload from the durable-extension: " + str(response)
            raise Exception(ex_msg)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.durable_functions.constants import ORCHESTRATION_TRIGGER, \
    ACTIVITY_TRIGGER, ENTITY_TRIGGER, DURABLE_CLIENT
from azure.functions.decorators.core import Trigger, InputBinding


class OrchestrationTrigger(Trigger):
    """OrchestrationTrigger.

    Trigger representing an Orchestration Function.
    """

    @staticmethod
    def get_binding_name() -> str:
        """Get the name of this trigger, as a string.

        Returns
        -------
        str
            The string representation of this trigger.
        """
        return ORCHESTRATION_TRIGGER

    def __init__(self,
                 name: str,
                 orchestration: Optional[str] = None,
                 ) -> None:
        self.orchestration = orchestration
        super().__init__(name=name)


class ActivityTrigger(Trigger):
    """ActivityTrigger.

    Trigger representing a Durable Functions Activity.
    """

    @staticmethod
    def get_binding_name() -> str:
        """Get the name of this trigger, as a string.

        Returns
        -------
        str
            The string representation of this trigger.
        """
        return ACTIVITY_TRIGGER

    def __init__(self,
                 name: str,
                 activity: Optional[str] = None,
                 ) -> None:
        self.activity = activity
        super().__init__(name=name)


class EntityTrigger(Trigger):
    """EntityTrigger.

    Trigger representing an Entity Function.
    """

    @staticmethod
    def get_binding_name() -> str:
        """Get the name of this trigger, as a string.

        Returns
        -------
        str
            The string representation of this trigger.
        """
        return ENTITY_TRIGGER

    def __init__(self,
                 name: str,
                 entity_name: Optional[str] = None,
                 ) -> None:
        self.entity_name = entity_name
        super().__init__(name=name)


class DurableClient(InputBinding):
    """DurableClient.

    Binding representing a Durable-client object.
    """

    @staticmethod
    def get_binding_name() -> str:
        """Get the name of this Binding, as a string.

        Returns
        -------
        str
            The string representation of this binding.
        """
        return DURABLE_CLIENT

    def __init__(self,
                 name: str,
                 task_hub: Optional[str] = None,
                 connection_name: Optional[str] = None
                 ) -> None:
        self.task_hub = task_hub
        self.connection_name = connection_name
        super().__init__(name=name)



import pytest
import json

import azure.durable_functions as df
import azure.functions as func
from tests.test_utils.constants import RPC_BASE_URL
from azure.durable_functions.models.DurableOrchestrationBindings import \
    DurableOrchestrationBindings

TASK_HUB_NAME = "DurableFunctionsHub"
BASE_URL = "http://localhost:7071/runtime/webhooks/durabletask"
AUTH_CODE = "iDFeaQCSAIuXoodl6/w3rdvHZ6Nl7yJwRrHfeInNWDJjuiunhxk8dQ=="


def get_binding_string():
    binding = {
        "taskHubName": TASK_HUB_NAME,
        "creationUrls": {
            "createNewInstancePostUri": f"{BASE_URL}/orchestrators/"
                                        "{functionName}[/{instanceId}]?code="
                                        f"{AUTH_CODE}",
            "createAndWaitOnNewInstancePostUri": f"{BASE_URL}/orchestrators/"
                                                 "{functionName}[/{instanceId}]?timeout="
                                                 "{timeoutInSeconds}&pollingInterval="
                                                 "{intervalInSeconds}&code="
                                                 f"{AUTH_CODE}"
        },
        "managementUrls": {
            "id": "INSTANCEID",
            "statusQueryGetUri": f"{BASE_URL}/instances/INSTANCEID?taskHub=DurableFunctionsHub&"
                                 f"connection=Storage&code={AUTH_CODE}",
            "sendEventPostUri": f"{BASE_URL}/instances/INSTANCEID/raiseEvent/"
                                "{eventName}?taskHub="
                                f"{TASK_HUB_NAME}&connection=Storage&code={AUTH_CODE}",
            "terminatePostUri": f"{BASE_URL}/instances/INSTANCEID/terminate?reason="
                                "{text}&taskHub="
                                f"{TASK_HUB_NAME}&connection=Storage&code={AUTH_CODE}",
            "rewindPostUri": f"{BASE_URL}/instances/INSTANCEID/rewind?reason="
                             "{text}&taskHub="
                             f"{TASK_HUB_NAME}&connection=Storage&code={AUTH_CODE}",
            "purgeHistoryDeleteUri": f"{BASE_URL}/instances/INSTANCEID?taskHub="
                                     f"{TASK_HUB_NAME}&connection=Storage&code={AUTH_CODE}"
        },
        "rpcBaseUrl": RPC_BASE_URL
    }
    binding_string = json.dumps(binding)

    binding_string = replace_stand_in_bits(binding_string)
    return binding_string


@pytest.fixture()
def binding_string():
    return get_binding_string()


@pytest.fixture()
def binding_info():
    binding = DurableOrchestrationBindings.from_json(get_binding_string())
    return binding


def replace_stand_in_bits(binding_string):
    binding_string = binding_string.replace("TASK_HUB_NAME", TASK_HUB_NAME)
    binding_string = binding_string.replace("BASE_URL", BASE_URL)
    binding_string = binding_string.replace("AUTH_CODE", AUTH_CODE)
    return binding_string

@pytest.fixture()
def app():
    app = df.DFApp(http_auth_level=func.AuthLevel.ANONYMOUS)
    return app



import logging

from azure.durable_functions import DurableOrchestrationClient
import azure.functions as func


async def main(req: func.HttpRequest, starter: str, message):
    """This function starts up the orchestrator from an HTTP endpoint

    starter: str
        A JSON-formatted string describing the orchestration context

    message:
        An azure functions http output binding, it enables us to establish
        an http response.

    Parameters
    ----------
    req: func.HttpRequest
        An HTTP Request object, it can be used to parse URL
        parameters.
    """


    function_name = req.route_params.get('functionName')
    logging.info(starter)
    client = DurableOrchestrationClient(starter)
    instance_id = await client.start_new(function_name)
    response = client.create_check_status_response(req, instance_id)
    message.set(response)



class SerializableClass(object):
    """Example serializable class.

    For a custom class to be serializable in
    Python Durable Functions, we require that
    it include both `to_json` and `from_json`
    a `@staticmethod`s for serializing to JSON
    and back respectively. These get called
    internally by the framework.
    """

    def __init__(self, name: str):
        """Construct the class.

        Parameters
        ----------
        number: int
            A number to encapsulate
        """
        self.name = name

    @staticmethod
    def to_json(obj: object) -> str:
        """Serialize a `SerializableClass` instance into a JSON string.

        Parameters
        ----------
        obj: SerializableClass
            The object to serialize

        Returns
        -------
        json_str: str
            A JSON-encoding of `obj`
        """
        return obj.name

    @staticmethod
    def from_json(json_str: str) -> object:
        """De-serialize a JSON string to a `SerializableClass` instance.

        It assumes that the JSON string was generated via
        `SerializableClass.to_json`

        Parameters
        ----------
        json_str: str
            The JSON-encoding of a `SerializableClass` instance

        Returns
        -------
        obj: SerializableClass
            A SerializableClass instance, de-serialized from `json_str`
        """
        obj = SerializableClass(json_str)
        return obj



# This function is not intended to be invoked directly. Instead it will be
# triggered by an HTTP starter function.
# Before running this sample, please:
# - create a Durable activity function (default name is "Hello")
# - create a Durable HTTP starter function
# - add azure-functions-durable to requirements.txt
# - run pip install -r requirements.txt

import logging
import json

import azure.functions as func
import azure.durable_functions as df


def orchestrator_function(context: df.DurableOrchestrationContext):
    input_ = context.get_input()
    result1 = yield context.call_activity('Hello', input_)
    return result1

main = df.Orchestrator.create(orchestrator_function)


import os
import logging
import time

from azureml.core.authentication import ServicePrincipalAuthentication

_TENANT_ID_ENV_NAME = "TenantId"
_SERVICE_PRINCIPAL_ID_ENV_NAME = "ServicePrincipalId"
_SERVICE_PRINCIPAL_SECRET_ENV_NAME = "ServicePrincipalSecret"


def get_service_principal_auth():
    tenant_id = os.environ[_TENANT_ID_ENV_NAME]
    service_principal_id = os.environ[_SERVICE_PRINCIPAL_ID_ENV_NAME]
    service_principal_password = os.environ[_SERVICE_PRINCIPAL_SECRET_ENV_NAME]

    svc_pr = ServicePrincipalAuthentication(
        tenant_id=tenant_id,
        service_principal_id=service_principal_id,
        service_principal_password=service_principal_password)

    return svc_pr


def get_access_token():
    start_time = time.time()

    svc_pr = get_service_principal_auth()
    aad_token = svc_pr.get_authentication_header()

    end_time = time.time()

    logging.info('Get Access Token Time: %s seconds', end_time - start_time)
    return aad_token



from azure.durable_functions.models.actions.Action import Action
from typing import Any, Dict, Optional


class NoOpAction(Action):
    """A no-op action, for anonymous tasks only."""

    def __init__(self, metadata: Optional[str] = None):
        """Create a NoOpAction object.

        This is an internal-only action class used to represent cases when intermediate
        tasks are used to implement some API. For example, in -WithRetry APIs, intermediate
        timers are created. We create this NoOp action to track those the backing actions
        of those tasks, which is necessary because we mimic the DF-internal replay algorithm.

        Parameters
        ----------
        metadata : Optional[str]
            Used for internal debugging: metadata about the action being represented.
        """
        self.metadata = metadata

    def action_type(self) -> int:
        """Get the type of action this class represents."""
        raise Exception("Attempted to get action type of an anonymous Action")

    def to_json(self) -> Dict[str, Any]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        raise Exception("Attempted to convert an anonymous Action to JSON")



"""Base module for the Python Durable functions.

Exposes the different API components intended for public consumption
"""
from .orchestrator import Orchestrator
from .entity import Entity
from .models.utils.entity_utils import EntityId
from .models.DurableOrchestrationClient import DurableOrchestrationClient
from .models.OrchestrationRuntimeStatus import OrchestrationRuntimeStatus
from .models.DurableOrchestrationContext import DurableOrchestrationContext
from .models.DurableEntityContext import DurableEntityContext
from .models.RetryOptions import RetryOptions
from .models.TokenSource import ManagedIdentityTokenSource
import json
from pathlib import Path
import sys
import warnings


def validate_extension_bundles():
    """Raise a warning if host.json contains bundle-range V1.

    Effects
    ------
        Warning: Warning prompting the user to update to bundles V2
    """
    # No need to validate if we're running tests
    if "pytest" in sys.modules:
        return

    host_path = "host.json"
    bundles_key = "extensionBundle"
    version_key = "version"
    host_file = Path(host_path)

    if not host_file.exists():
        # If it doesn't exist, we ignore it
        return

    with open(host_path) as f:
        host_settings = json.loads(f.read())
        try:
            version_range = host_settings[bundles_key][version_key]
        except Exception:
            # If bundle info is not available, we ignore it.
            # For example: it's possible the user is using a manual extension install
            return
        # We do a best-effort attempt to detect bundles V1
        # This is the string hard-coded into the bundles V1 template in VSCode
        if version_range == "[1.*, 2.0.0)":
            message = "Your application is currently configured to use Extension Bundles V1."\
                " Durable Functions for Python works best with Bundles V2,"\
                " which provides additional features like Durable Entities, better performance,"\
                " and is actively being developed."\
                " Please update to Bundles V2 in your `host.json`."\
                " You can set extensionBundles version to be: [2.*, 3.0.0)"
            warnings.warn(message)


# Validate that users are not in extension bundles V1
validate_extension_bundles()

__all__ = [
    'Orchestrator',
    'Entity',
    'EntityId',
    'DurableOrchestrationClient',
    'DurableEntityContext',
    'DurableOrchestrationContext',
    'ManagedIdentityTokenSource',
    'OrchestrationRuntimeStatus',
    'RetryOptions'
]

try:
    # disabling linter on this line because it fails to recognize the conditional export
    from .decorators.durable_app import (DFApp, Blueprint) # noqa
    __all__.append('DFApp')
    __all__.append('Blueprint')
except ModuleNotFoundError:
    pass



import logging

def main(name):
    """Activity function performing a specific step in the chain
    
    Parameters
    ----------
    name : str
        Name of the item to be hello'ed at 
    
    Returns
    -------
    str
        Returns a welcome string
    """
    logging.warning(f"Activity Triggered: {name}")
    return name

    


from azure.durable_functions.models.actions.CompoundAction import CompoundAction
from .ActionType import ActionType


class WhenAnyAction(CompoundAction):
    """Defines the structure of the WhenAll Action object.

    Provides the information needed by the durable extension to be able to invoke WhenAll tasks.
    """

    @property
    def action_type(self) -> int:
        """Get the type of action this class represents."""
        return ActionType.WHEN_ANY



from enum import IntEnum


class HistoryEventType(IntEnum):
    """Defines the different types of history events being communicated."""

    EXECUTION_STARTED = 0
    EXECUTION_COMPLETED = 1
    EXECUTION_FAILED = 2
    EXECUTION_TERMINATED = 3
    TASK_SCHEDULED = 4
    TASK_COMPLETED = 5
    TASK_FAILED = 6
    SUB_ORCHESTRATION_INSTANCE_CREATED = 7
    SUB_ORCHESTRATION_INSTANCE_COMPLETED = 8
    SUB_ORCHESTRATION_INSTANCE_FAILED = 9
    TIMER_CREATED = 10
    TIMER_FIRED = 11
    ORCHESTRATOR_STARTED = 12
    ORCHESTRATOR_COMPLETED = 13
    EVENT_SENT = 14
    EVENT_RAISED = 15
    CONTINUE_AS_NEW = 16
    GENERIC_EVENT = 17
    HISTORY_STATE = 18



from typing import Dict, Union

from .utils.json_utils import add_attrib


class RetryOptions:
    """Retry Options.

    Defines retry policies that can be passed as parameters to various
    operations.
    """

    def __init__(
            self,
            first_retry_interval_in_milliseconds: int,
            max_number_of_attempts: int):
        self._first_retry_interval_in_milliseconds: int = \
            first_retry_interval_in_milliseconds
        self._max_number_of_attempts: int = max_number_of_attempts

        if self._first_retry_interval_in_milliseconds <= 0:
            raise ValueError("first_retry_interval_in_milliseconds value"
                             "must be greater than 0.")

    @property
    def first_retry_interval_in_milliseconds(self) -> int:
        """Get the first retry interval (ms).

        Must be greater than 0

        Returns
        -------
        int
            The value indicating the first retry interval
        """
        return self._first_retry_interval_in_milliseconds

    @property
    def max_number_of_attempts(self) -> int:
        """Get Max Number of Attempts.

        Returns
        -------
        int
            Value indicating the max number of attempts to retry
        """
        return self._max_number_of_attempts

    def to_json(self) -> Dict[str, Union[str, int]]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Union[str, int]] = {}

        add_attrib(
            json_dict,
            self,
            'first_retry_interval_in_milliseconds',
            'firstRetryIntervalInMilliseconds')
        add_attrib(
            json_dict,
            self,
            'max_number_of_attempts',
            'maxNumberOfAttempts')
        return json_dict



from typing import Any, Dict, Union

from .ActionType import ActionType
from .Action import Action
from ..utils.json_utils import add_attrib, add_datetime_attrib
import datetime


class CreateTimerAction(Action):
    """Defines the structure of the Create Timer object.

    Returns
    -------
        Information needed by durable extension to schedule the activity

    Raises
    ------
    ValueError
        if the event fired is not of valid datetime object
    """

    def __init__(self, fire_at: datetime.datetime, is_cancelled: bool = False):
        self._action_type: ActionType = ActionType.CREATE_TIMER
        self.fire_at: datetime.datetime = fire_at
        self.is_cancelled: bool = is_cancelled

        if not isinstance(self.fire_at, datetime.date):
            raise ValueError("fireAt: Expected valid datetime object but got ", self.fire_at)

    def to_json(self) -> Dict[str, Any]:
        """
        Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Union[int, str]] = {}
        add_attrib(json_dict, self, 'action_type', 'actionType')
        add_datetime_attrib(json_dict, self, 'fire_at', 'fireAt')
        add_attrib(json_dict, self, 'is_cancelled', 'isCanceled')
        return json_dict

    @property
    def action_type(self) -> int:
        """Get the type of action this class represents."""
        return ActionType.CREATE_TIMER



from azure.durable_functions.models.ReplaySchema import ReplaySchema
from tests.test_utils.ContextBuilder import ContextBuilder
from .orchestrator_test_utils \
    import get_orchestration_state_result, assert_orchestration_state_equals, assert_valid_schema
from azure.durable_functions.models.OrchestratorState import OrchestratorState

def base_expected_state(output=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema.value)

def generator_function(context):
    return False

def test_serialization_of_False():
    """Test that an orchestrator can return False."""

    context_builder = ContextBuilder("serialize False")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state(output=False)

    expected_state._is_done = True
    expected = expected_state.to_json()

    # Since we're essentially testing the `to_json` functionality,
    # we explicitely ensure that the output is set
    expected["output"] = False

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


from typing import Dict, Any
import json


class ResponseMessage:
    """ResponseMessage.

    Specifies the response of an entity, as processed by the durable-extension.
    """

    def __init__(self, result: str, is_exception: bool = False):
        """Instantiate a ResponseMessage.

        Specifies the response of an entity, as processed by the durable-extension.

        Parameters
        ----------
        result: str
            The result provided by the entity
        """
        # The time-out case seems to be handled by the Functions-Host, so
        # its result is not doubly-serialized. In this branch, we compensate
        # for this by re-serializing the payload.
        if result.strip().startswith("Timeout value of"):
            is_exception = True
            result = json.dumps(result)

        self.result = result
        self.is_exception = is_exception
        # TODO: JS has an additional exceptionType field, but does not use it

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> 'ResponseMessage':
        """Instantiate a ResponseMessage from a dict of the JSON-response by the extension.

        Parameters
        ----------
        d: Dict[str, Any]
            The dictionary parsed from the JSON-response by the durable-extension

        Returns
        -------
        ResponseMessage:
            The ResponseMessage built from the provided dictionary
        """
        is_error = "exceptionType" in d.keys()
        result = cls(d["result"], is_error)
        return result



schema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "isDone": {"type": "boolean"},
        "output": {},
        "error": {"type": "string"},
        "customStatus": { 
                        "anyOf": [
                        {
                            "type": "object",
                        },{
                            "type": "string",
                        }] },
        "actions": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "functionName": {"type": "string"},
                        "actionType": {"type": "number"},
                        "input": {},
                        "retryOptions": {
                            "type": "object",
                            "properties": {
                                "firstRetryIntervalInMilliseconds": {
                                    "type": "number",
                                    "minimum": 1},
                                "maxNumberOfAttempts": {"type": "number"}
                            },
                            "required":
                                ["firstRetryIntervalInMilliseconds", "maxNumberOfAttempts"],
                            "additionalProperties": False
                        },
                        "httpRequest": {
                            "type": "object",
                            "properties": {
                                "method": {"type": "string"},
                                "uri": {"type": "string"},
                                "content": {},
                                "headers": {},
                                "tokenSource": {
                                    "type": "object",
                                    "properties": {
                                        "resource": {"type": "string"}
                                    },
                                    "required": ["resource"],
                                    "additionalProperties": False
                                }
                            },
                            "required":
                                ["method", "uri"],
                            "additionalProperties": False
                        }
                    },
                    "required": ["actionType"],
                    "additionalProperties": False
                }
            }
        }
    },
    "required": ["isDone"],
    "additionalProperties": False
}



from azure.durable_functions.models.ReplaySchema import ReplaySchema
from azure.durable_functions.models.actions.WhenAnyAction import WhenAnyAction
from tests.test_utils.ContextBuilder import ContextBuilder
from .orchestrator_test_utils \
    import get_orchestration_state_result, assert_orchestration_state_equals, assert_valid_schema
from azure.durable_functions.models.actions.CreateTimerAction import CreateTimerAction
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.constants import DATETIME_STRING_FORMAT
from datetime import datetime, timedelta, timezone


def base_expected_state(output=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema)

def add_timer_fired_events(context_builder: ContextBuilder, id_: int, timestamp: str):
    fire_at: str = context_builder.add_timer_created_event(id_, timestamp)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_timer_fired_event(id_=id_, fire_at=fire_at)

def generator_function(context):

    # Create a timezone aware datetime object, just like a normal
    # call to `context.current_utc_datetime` would create
    timestamp = "2020-07-23T21:56:54.936700Z"
    fire_at = datetime.strptime(timestamp, DATETIME_STRING_FORMAT)
    fire_at = fire_at.replace(tzinfo=timezone.utc)

    yield context.create_timer(fire_at)
    return "Done!"

def generator_function_timer_can_be_cancelled(context):
    time_limit1 = context.current_utc_datetime + timedelta(minutes=5)
    timer_task1 = context.create_timer(time_limit1)

    time_limit2 = context.current_utc_datetime + timedelta(minutes=10)
    timer_task2 = context.create_timer(time_limit2)

    winner = yield context.task_any([timer_task1, timer_task2])
    if winner == timer_task1:
        timer_task2.cancel()
        return "Done!"
    else:
        raise Exception("timer task 1 should complete before timer task 2")

def add_timer_action(state: OrchestratorState, fire_at: datetime):
    action = CreateTimerAction(fire_at=fire_at)
    state._actions.append([action])

def test_timers_comparison_with_relaxed_precision():
    """Test if that two `datetime` different but equivalent
    serializations of timer deadlines are found to be equivalent.

    The Durable Extension may sometimes drop redundant zeroes on
    a datetime object. For instance, the date
        2020-07-23T21:56:54.936700Z
    may get transformed into
        2020-07-23T21:56:54.9367Z
    This test ensures that dropping redundant zeroes does not affect
    our ability to recognize that a timer has been fired.
    """

    # equivalent to 2020-07-23T21:56:54.936700Z
    relaxed_timestamp = "2020-07-23T21:56:54.9367Z"
    fire_at = datetime.strptime(relaxed_timestamp, DATETIME_STRING_FORMAT)

    context_builder = ContextBuilder("relaxed precision")
    add_timer_fired_events(context_builder, 0, relaxed_timestamp)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state(output='Done!')
    add_timer_action(expected_state, fire_at)

    expected_state._is_done = True
    expected = expected_state.to_json()

    #assert_valid_schema(result)
    # TODO: getting the following error when validating the schema
    # "Additional properties are not allowed ('fireAt', 'isCanceled' were unexpected)">
    assert_orchestration_state_equals(expected, result)

def test_timers_can_be_cancelled():

    context_builder = ContextBuilder("test_timers_can_be_cancelled")
    fire_at1 = context_builder.current_datetime + timedelta(minutes=5)
    fire_at2 = context_builder.current_datetime + timedelta(minutes=10)
    add_timer_fired_events(context_builder, 0, str(fire_at1))
    add_timer_fired_events(context_builder, 1, str(fire_at2))

    result = get_orchestration_state_result(
        context_builder, generator_function_timer_can_be_cancelled)

    expected_state = base_expected_state(output='Done!')
    expected_state._actions.append(
        [CreateTimerAction(fire_at=fire_at1), CreateTimerAction(fire_at=fire_at2, is_cancelled=True)])

    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_orchestration_state_equals(expected, result)
    assert result["actions"][0][1]["isCanceled"]

def test_timers_can_be_cancelled_replayV2():

    context_builder = ContextBuilder("test_timers_can_be_cancelled", replay_schema=ReplaySchema.V2)
    fire_at1 = context_builder.current_datetime + timedelta(minutes=5)
    fire_at2 = context_builder.current_datetime + timedelta(minutes=10)
    add_timer_fired_events(context_builder, 0, str(fire_at1))
    add_timer_fired_events(context_builder, 1, str(fire_at2))

    result = get_orchestration_state_result(
        context_builder, generator_function_timer_can_be_cancelled)

    expected_state = base_expected_state(output='Done!', replay_schema=ReplaySchema.V2)
    expected_state._actions = [
        [WhenAnyAction(
            [CreateTimerAction(fire_at=fire_at1), CreateTimerAction(fire_at=fire_at2, is_cancelled=True)]
        )]
    ]

    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_orchestration_state_equals(expected, result)
    assert result["actions"][0][0]['compoundActions'][1]["isCanceled"]


from tests.test_utils.ContextBuilder import ContextBuilder
from tests.test_utils.testClasses import SerializableClass
from azure.durable_functions.models.RetryOptions import RetryOptions
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.models.DurableOrchestrationContext import DurableOrchestrationContext
from .orchestrator_test_utils import get_orchestration_state_result
from typing import List, Tuple
from datetime import datetime

RETRY_OPTIONS = RetryOptions(5000, 2)
REASONS = "Stuff"
DETAILS = "Things"
RESULT_PREFIX = "Hello "
CITIES = ["Tokyo", "Seattle", "London"]

def generator_function(context: DurableOrchestrationContext):
    """Orchestrator function for testing retry'ing semantics

    Parameters
    ----------
    context: DurableOrchestrationContext
        Durable orchestration context, exposes the Durable API

    Returns
    -------
    List[str]:
        Output of activities, a list of hello'd cities
    """

    outputs = []

    retry_options = RETRY_OPTIONS
    task1 = yield context.call_activity_with_retry(
        "Hello", retry_options, "Tokyo")
    task2 = yield context.call_activity_with_retry(
        "Hello",  retry_options, "Seattle")
    task3 = yield context.call_activity_with_retry(
        "Hello",  retry_options, "London")

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs


def generator_function_with_serialization(context: DurableOrchestrationContext):
    """Orchestrator function for testing retry'ing with serializable input arguments.

    Parameters
    ----------
    context: DurableOrchestrationContext
        Durable orchestration context, exposes the Durable API

    Returns
    -------
    List[str]:
        Output of activities, a list of hello'd cities
    """

    outputs = []

    retry_options = RETRY_OPTIONS
    task1 = yield context.call_activity_with_retry(
        "Hello", retry_options, SerializableClass("Tokyo"))
    task2 = yield context.call_activity_with_retry(
        "Hello",  retry_options, SerializableClass("Seatlle"))
    task3 = yield context.call_activity_with_retry(
        "Hello",  retry_options, SerializableClass("London"))

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs


def get_context_with_retries_and_corrupted_completion() -> ContextBuilder:
    """Get a ContextBuilder whose history contains a late completion event
    for an event that already failed.

    Returns
    -------
    ContextBuilder:
        The context whose history contains the requested event sequence.
    """
    context = get_context_with_retries()
    context.add_orchestrator_started_event()
    context.add_task_completed_event(id_=0, result="'Do not pick me up'")
    context.add_orchestrator_completed_event()
    return context

def get_context_with_retries(will_fail: bool=False) -> ContextBuilder:
    """Get a ContextBuilder whose history contains retried events.

    Parameters
    ----------
    will_fail: (bool, optional)
        If set to true, returns a context with a history where the orchestrator fails.
        If false, returns a context with a history where events fail but eventually complete.
        Defaults to False.

    Returns
    -------
    ContextBuilder:
        The context whose history contains the requested event sequence.
    """
    context = ContextBuilder()
    num_activities = len(CITIES)

    def _schedule_events(context: ContextBuilder, id_counter: int) -> Tuple[ContextBuilder, int, List[int]]:
        """Add scheduled events to the context.

        Parameters
        ----------
        context: ContextBuilder
            Orchestration context mock, to which we'll add the event completion events
        id_counter: int
            The current event counter

        Returns
        -------
        Tuple[ContextBuilder, int, List[int]]:
            The updated context, the updated counter, a list of event IDs for each scheduled event
        """
        id_counter = id_counter + 1
        context.add_task_scheduled_event(name='Hello', id_=id_counter)
        return context, id_counter

    def _fail_events(context: ContextBuilder, id_counter: int) -> Tuple[ContextBuilder, int]:
        """Add event failed to the context.

        Parameters
        ----------
        context: ContextBuilder
            Orchestration context mock, to which we'll add the event completion events
        id_counter: int
            The current event counter

        Returns
        -------
        Tuple[ContextBuilder, int]:
            The updated context, the updated id_counter
        """
        context.add_orchestrator_started_event()
        context.add_task_failed_event(
            id_=id_counter, reason=REASONS, details=DETAILS)
        return context, id_counter

    def _schedule_timers(context: ContextBuilder, id_counter: int) -> Tuple[ContextBuilder, int, List[datetime]]:
        """Add timer created events to the context.

        Parameters
        ----------
        context: ContextBuilder
            Orchestration context mock, to which we'll add the event completion events
        id_counter: int
            The current event counter

        Returns
        -------
        Tuple[ContextBuilder, int, List[datetime]]:
            The updated context, the updated counter, a list of timer deadlines
        """
        id_counter = id_counter + 1
        deadlines: List[datetime] = []
        deadlines.append((id_counter, context.add_timer_created_event(id_counter)))
        return context, id_counter, deadlines
    
    def _fire_timer(context: ContextBuilder, id_counter: int, deadlines: List[datetime]) -> Tuple[ContextBuilder, int]:
        """Add timer fired events to the context.

        Parameters
        ----------
        context: ContextBuilder
            Orchestration context mock, to which we'll add the event completion events
        id_counter: int
            The current event counter
        deadlines: List[datetime]
            List of dates at which to fire the timers

        Returns
        -------
        Tuple[ContextBuilder, int]:
            The updated context, the updated id_counter
        """
        for id_, fire_at in deadlines:
            context.add_timer_fired_event(id_=id_, fire_at=fire_at)
        return context, id_counter

    def _complete_event(context: ContextBuilder, id_counter: int, city:str) -> Tuple[ContextBuilder, int]:
        """Add event / task completions to the context.

        Parameters
        ----------
        context: ContextBuilder
            Orchestration context mock, to which we'll add the event completion events
        id_counter: int
            The current event counter

        Returns
        -------
        Tuple[ContextBuilder, int]
            The updated context, the updated id_counter
        """
        result = f"\"{RESULT_PREFIX}{city}\""
        context.add_task_completed_event(id_=id_counter, result=result)
        return context, id_counter


    id_counter = -1

    for city in CITIES:
        # Schedule the events
        context, id_counter = _schedule_events(context, id_counter)
        context.add_orchestrator_completed_event()

        # Record failures, schedule timers
        context, id_counter = _fail_events(context, id_counter)
        context, id_counter, deadlines = _schedule_timers(context, id_counter)
        context.add_orchestrator_completed_event()

        # Fire timers, re-schedule events
        context.add_orchestrator_started_event()
        context, id_counter = _fire_timer(context, id_counter, deadlines)
        context, id_counter = _schedule_events(context, id_counter)
        context.add_orchestrator_completed_event()

        context.add_orchestrator_started_event()

        # Either complete the event or, if we want all failed events, then
        # fail the events, schedule timer, and fire time.
        if will_fail:
            context, id_counter = _fail_events(context, id_counter)
            context, id_counter, deadlines = _schedule_timers(context, id_counter)
            context.add_orchestrator_completed_event()

            context.add_orchestrator_started_event()
            context, id_counter = _fire_timer(context, id_counter, deadlines)
        else:
            context, id_counter = _complete_event(context, id_counter, city)

        context.add_orchestrator_completed_event()
    return context

def test_redundant_completion_doesnt_get_processed():
    """Tests that our implementation processes the state array
    sequentially, which previous implementations did not guarantee. In this test,
    we add a completion event for a task that was cancelled, meaning that it failed and got
    re-scheduled. Older implementations would pick up this completion event and cause
    non-determinism.
    """
    context_1 = get_context_with_retries()
    context_2 = get_context_with_retries_and_corrupted_completion()

    result_1 = get_orchestration_state_result(
        context_1, generator_function)

    result_2 = get_orchestration_state_result(
        context_2, generator_function)

    assert "output" in result_1
    assert "output" in result_2
    assert result_1["output"] == result_2["output"]


def test_failed_tasks_do_not_hang_orchestrator():
    """Tests that our implementation correctly handles up re-scheduled events,
    which previous implementations failed to correctly handle. """
    context = get_context_with_retries()

    result = get_orchestration_state_result(
        context, generator_function)

    expected_output = list(map(lambda x: RESULT_PREFIX + x, CITIES))
    assert "output" in result
    assert result["output"] == expected_output

def test_retries_can_fail():
    """Tests the code path where a retry'ed Task fails"""
    context = get_context_with_retries(will_fail=True)

    try:
        result = get_orchestration_state_result(
            context, generator_function)
        # We expected an exception
        assert False
    except Exception as e:
        error_label = "\n\n$OutOfProcData$:"
        error_str = str(e)

        error_msg = f"{REASONS} \n {DETAILS}"
        
        expected_error_str = f"{error_msg}{error_label}"
        assert str.startswith(error_str, expected_error_str)

def test_retries_with_serializable_input():
    # Tests that retried tasks work with serialized input classes.
    context = get_context_with_retries()

    result_1 = get_orchestration_state_result(
        context, generator_function)

    result_2 = get_orchestration_state_result(
        context, generator_function_with_serialization)

    assert "output" in result_1
    assert "output" in result_2
    assert result_1["output"] == result_2["output"]


from azure.durable_functions.models.actions.NoOpAction import NoOpAction
from azure.durable_functions.models.actions.CompoundAction import CompoundAction
from azure.durable_functions.models.RetryOptions import RetryOptions
from azure.durable_functions.models.ReplaySchema import ReplaySchema
from azure.durable_functions.models.actions.Action import Action
from azure.durable_functions.models.actions.WhenAnyAction import WhenAnyAction
from azure.durable_functions.models.actions.WhenAllAction import WhenAllAction
from azure.durable_functions.models.actions.CreateTimerAction import CreateTimerAction

import enum
from typing import Any, List, Optional, Set, Type, Union


class TaskState(enum.Enum):
    """The possible states that a Task can be in."""

    RUNNING = 0
    SUCCEEDED = 1
    FAILED = 2


class TaskBase:
    """The base class of all Tasks.

    Contains shared logic that drives all of its sub-classes. Should never be
    instantiated on its own.
    """

    def __init__(self, id_: Union[int, str], actions: Union[List[Action], Action]):
        """Initialize the TaskBase.

        Parameters
        ----------
        id_ : int
            An ID for the task
        actions : List[Any]
            The list of DF actions representing this Task.
            Needed for reconstruction in the extension.
        """
        self.id: Union[int, str] = id_
        self.state = TaskState.RUNNING
        self.parent: Optional[CompoundTask] = None
        self._api_name: str

        api_action: Union[Action, Type[CompoundAction]]
        if isinstance(actions, list):
            if len(actions) == 1:
                api_action = actions[0]
            else:
                api_action = CompoundAction
        else:
            api_action = actions

        self._api_name = api_action.__class__.__name__

        self.result: Any = None
        self.action_repr: Union[List[Action], Action] = actions
        self.is_played = False
        self._is_scheduled_flag = False

    @property
    def _is_scheduled(self) -> bool:
        return self._is_scheduled_flag

    def _set_is_scheduled(self, is_scheduled: bool):
        self._is_scheduled_flag = is_scheduled

    @property
    def is_completed(self) -> bool:
        """Get indicator of whether the task completed.

        Note that completion is not equivalent to success.
        """
        return not (self.state is TaskState.RUNNING)

    def set_is_played(self, is_played: bool):
        """Set the is_played flag for the Task.

        Needed for updating the orchestrator's is_replaying flag.

        Parameters
        ----------
        is_played : bool
            Whether the latest event for this Task has been played before.
        """
        self.is_played = is_played

    def change_state(self, state: TaskState):
        """Transition a running Task to a terminal state: success or failure.

        Parameters
        ----------
        state : TaskState
            The terminal state to assign to this Task

        Raises
        ------
        Exception
            When the input state is RUNNING
        """
        if state is TaskState.RUNNING:
            raise Exception("Cannot change Task to the RUNNING state.")
        self.state = state

    def set_value(self, is_error: bool, value: Any):
        """Set the value of this Task: either an exception of a result.

        Parameters
        ----------
        is_error : bool
            Whether the value represents an exception of a result.
        value : Any
            The value of this Task

        Raises
        ------
        Exception
            When the Task failed but its value was not an Exception
        """
        new_state = self.state
        if is_error:
            if not isinstance(value, Exception):
                if not (isinstance(value, TaskBase) and isinstance(value.result, Exception)):
                    err_message = f"Task ID {self.id} failed but it's value was not an Exception"
                    raise Exception(err_message)
            new_state = TaskState.FAILED
        else:
            new_state = TaskState.SUCCEEDED
        self.change_state(new_state)
        self.result = value
        self.propagate()

    def propagate(self):
        """Notify parent Task of this Task's state change."""
        has_completed = not (self.state is TaskState.RUNNING)
        has_parent = not (self.parent is None)
        if has_completed and has_parent:
            self.parent.handle_completion(self)


class CompoundTask(TaskBase):
    """A Task of Tasks.

    Contains shared logic that drives all of its sub-classes.
    Should never be instantiated on its own.
    """

    def __init__(self, tasks: List[TaskBase], compound_action_constructor=None):
        """Instantiate CompoundTask attributes.

        Parameters
        ----------
        tasks : List[Task]
            The children/sub-tasks of this Task
        compound_action_constructor : Union[WhenAllAction, WhenAnyAction, None]
            Either None or, a WhenAllAction or WhenAnyAction constructor.
            It is None when using the V1 replay protocol, where no Compound Action
            objects size and compound actions are represented as arrays of actions.
            It is not None when using the V2 replay protocol.
        """
        super().__init__(-1, [])
        child_actions = []
        for task in tasks:
            task.parent = self
            action_repr = task.action_repr
            if isinstance(action_repr, list):
                child_actions.extend(action_repr)
            else:
                if not task._is_scheduled:
                    child_actions.append(action_repr)
        if compound_action_constructor is None:
            self.action_repr = child_actions
        else:  # replay_schema is ReplaySchema.V2
            self.action_repr = compound_action_constructor(child_actions)
        self._first_error: Optional[Exception] = None
        self.pending_tasks: Set[TaskBase] = set(tasks)
        self.completed_tasks: List[TaskBase] = []
        self.children = tasks

        if len(self.children) == 0:
            self.state = TaskState.SUCCEEDED

        # Sub-tasks may have already completed, so we process them
        for child in self.children:
            if not (child.state is TaskState.RUNNING):
                self.handle_completion(child)

    @property
    def _is_scheduled(self) -> bool:
        return all([child._is_scheduled for child in self.children])

    def handle_completion(self, child: TaskBase):
        """Manage sub-task completion events.

        Parameters
        ----------
        child : TaskBase
            The sub-task that completed

        Raises
        ------
        Exception
            When the calling sub-task was not registered
            with this Task's pending sub-tasks.
        """
        try:
            self.pending_tasks.remove(child)
        except KeyError:
            raise Exception(
                f"Parent Task {self.id} does not have pending sub-task with ID {child.id}."
                f"This most likely means that Task {child.id} completed twice.")

        self.completed_tasks.append(child)
        self.set_is_played(child.is_played)
        self.try_set_value(child)

    def try_set_value(self, child: TaskBase):
        """Transition a CompoundTask to a terminal state and set its value.

        Should be implemented by sub-classes.

        Parameters
        ----------
        child : TaskBase
            A sub-task that just completed

        Raises
        ------
        NotImplementedError
            This method needs to be implemented by each subclass.
        """
        raise NotImplementedError


class AtomicTask(TaskBase):
    """A Task with no subtasks."""

    def _get_action(self) -> Action:
        action: Action
        if isinstance(self.action_repr, list):
            action = self.action_repr[0]
        else:
            action = self.action_repr
        return action


class TimerTask(AtomicTask):
    """A Timer Task."""

    def __init__(self, id_: Union[int, str], action: CreateTimerAction):
        super().__init__(id_, action)
        self.action_repr: Union[List[CreateTimerAction], CreateTimerAction]

    @property
    def is_cancelled(self) -> bool:
        """Check if the Timer is cancelled.

        Returns
        -------
        bool
            Returns whether a timer has been cancelled or not
        """
        action: CreateTimerAction = self._get_action()
        return action.is_cancelled

    def cancel(self):
        """Cancel a timer.

        Raises
        ------
        ValueError
            Raises an error if the task is already completed and an attempt is made to cancel it
        """
        if not self.is_completed:
            action: CreateTimerAction = self._get_action()
            action.is_cancelled = True
        else:
            raise ValueError("Cannot cancel a completed task.")


class WhenAllTask(CompoundTask):
    """A Task representing `when_all` scenarios."""

    def __init__(self, task: List[TaskBase], replay_schema: ReplaySchema):
        """Initialize a WhenAllTask.

        Parameters
        ----------
        task : List[Task]
            The list of child tasks
        replay_schema : ReplaySchema
            The ReplaySchema, which determines the inner action payload representation
        """
        compound_action_constructor = None
        if replay_schema is ReplaySchema.V2:
            compound_action_constructor = WhenAllAction
        super().__init__(task, compound_action_constructor)

    def try_set_value(self, child: TaskBase):
        """Transition a WhenAll Task to a terminal state and set its value.

        Parameters
        ----------
        child : TaskBase
            A sub-task that just completed
        """
        if child.state is TaskState.SUCCEEDED:
            # A WhenAll Task only completes when it has no pending tasks
            # i.e _when all_ of its children have completed
            if len(self.pending_tasks) == 0:
                results = list(map(lambda x: x.result, self.children))
                self.set_value(is_error=False, value=results)
        else:  # child.state is TaskState.FAILED:
            # a single error is sufficient to fail this task
            if self._first_error is None:
                self._first_error = child.result
                self.set_value(is_error=True, value=self._first_error)


class WhenAnyTask(CompoundTask):
    """A Task representing `when_any` scenarios."""

    def __init__(self, task: List[TaskBase], replay_schema: ReplaySchema):
        """Initialize a WhenAnyTask.

        Parameters
        ----------
        task : List[Task]
            The list of child tasks
        replay_schema : ReplaySchema
            The ReplaySchema, which determines the inner action payload representation
        """
        compound_action_constructor = None
        if replay_schema is ReplaySchema.V2:
            compound_action_constructor = WhenAnyAction
        super().__init__(task, compound_action_constructor)

    def try_set_value(self, child: TaskBase):
        """Transition a WhenAny Task to a terminal state and set its value.

        Parameters
        ----------
        child : TaskBase
            A sub-task that just completed
        """
        if self.state is TaskState.RUNNING:
            self.set_value(is_error=False, value=child)


class RetryAbleTask(WhenAllTask):
    """A Task representing `with_retry` scenarios.

    It inherits from WhenAllTask because retryable scenarios are Tasks
    with equivalent to WhenAll Tasks with dynamically increasing lists
    of children. At every failure, we add a Timer child and a Task child
    to the list of pending tasks.
    """

    def __init__(self, child: TaskBase, retry_options: RetryOptions, context):
        tasks = [child]
        super().__init__(tasks, context._replay_schema)

        self.retry_options = retry_options
        self.num_attempts = 1
        self.context = context
        self.actions = child.action_repr
        self.is_waiting_on_timer = False
        self.error = None

    @property
    def id_(self):
        """Obtain the task's ID.

        Since this is an internal-only abstraction, the task ID is represented
        by the ID of its inner/wrapped task _plus_ a suffix: "_retryable_proxy"

        Returns
        -------
        [type]
            [description]
        """
        return str(list(map(lambda x: x.id, self.children))) + "_retryable_proxy"

    def try_set_value(self, child: TaskBase):
        """Transition a Retryable Task to a terminal state and set its value.

        Parameters
        ----------
        child : TaskBase
            A sub-task that just completed
        """
        if self.is_waiting_on_timer:
            # timer fired, re-scheduling original task
            self.is_waiting_on_timer = False
            # As per DTFx semantics: we need to check the number of retires only after the final
            # timer has fired. This means we essentially have to wait for one "extra" timer after
            # the maximum number of attempts has been reached. Removing this extra timer will cause
            # stuck orchestrators as we need to be "in sync" with the replay logic of DTFx.
            if self.num_attempts >= self.retry_options.max_number_of_attempts:
                self.is_waiting_on_timer = True
                # we have reached the maximum number of attempts, set error
                self.set_value(is_error=True, value=self.error)
            else:
                rescheduled_task = self.context._generate_task(
                    action=NoOpAction("rescheduled task"), parent=self)
                self.pending_tasks.add(rescheduled_task)
                self.context._add_to_open_tasks(rescheduled_task)
                self.num_attempts += 1

            return
        if child.state is TaskState.SUCCEEDED:
            if len(self.pending_tasks) == 0:
                # if all pending tasks have completed,
                # and we have a successful child, then
                # we can set the Task's event
                self.set_value(is_error=False, value=child.result)

        else:  # child.state is TaskState.FAILED:
            # increase size of pending tasks by adding a timer task
            # when it completes, we'll retry the original task
            timer_task = self.context._generate_task(
                action=NoOpAction("-WithRetry timer"), parent=self)
            self.pending_tasks.add(timer_task)
            self.context._add_to_open_tasks(timer_task)
            self.is_waiting_on_timer = True
            self.error = child.result



import json
from typing import List, Any, Dict, Optional, Union

from azure.durable_functions.models.ReplaySchema import ReplaySchema

from .utils.json_utils import add_attrib
from azure.durable_functions.models.actions.Action import Action


class OrchestratorState:
    """Orchestration State.

    Used to communicate the state of the orchestration back to the durable
    extension
    """

    def __init__(self,
                 is_done: bool,
                 actions: List[List[Action]],
                 output: Any,
                 replay_schema: ReplaySchema,
                 error: str = None,
                 custom_status: Any = None):
        self._is_done: bool = is_done
        self._actions: List[List[Action]] = actions
        self._output: Any = output
        self._error: Optional[str] = error
        self._custom_status: Any = custom_status
        self._replay_schema: ReplaySchema = replay_schema

    @property
    def actions(self) -> List[List[Action]]:
        """Get the ordered list of async actions the orchestrator function should perform.

        This list is append-only; it must contain all scheduled async actions up to the latest
        requested work, even actions that have already been completed.

        Actions are grouped by execution. Each subsequent orchestrator execution should add a
        new array of action objects to the collection.
        """
        return self._actions

    @property
    def is_done(self) -> bool:
        """Get indicator of whether this is the last execution of this orchestrator instance.

        When this value is true, the Durable Functions extension will consider the orchestration
        instance completed and will attempt to return the output value.
        """
        return self._is_done

    @property
    def output(self):
        """Get the JSON-serializable value returned by the orchestrator instance completion.

        Optional.
        """
        return self._output

    @property
    def error(self) -> Optional[str]:
        """Get the error received when running the orchestration.

        Optional.
        """
        return self._error

    @property
    def custom_status(self):
        """Get the JSON-serializable value used by DurableOrchestrationContext.SetCustomStatus."""
        return self._custom_status

    @property
    def schema_version(self):
        """Get the Replay Schema represented in this OrchestratorState payload."""
        return self._replay_schema.value

    def to_json(self) -> Dict[str, Union[str, int]]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Union[str, int]] = {}
        add_attrib(json_dict, self, '_is_done', 'isDone')
        if self._replay_schema != ReplaySchema.V1:
            add_attrib(json_dict, self, 'schema_version', 'schemaVersion')
        self._add_actions(json_dict)
        if not (self._output is None):
            json_dict['output'] = self._output
        if self._error:
            json_dict['error'] = self._error
        if self._custom_status:
            json_dict['customStatus'] = self._custom_status
        return json_dict

    def _add_actions(self, json_dict):
        json_dict['actions'] = []
        for action_list in self._actions:
            action_result_list = []
            for action_obj in action_list:
                action_result_list.append(action_obj.to_json())
            json_dict['actions'].append(action_result_list)

    def to_json_string(self) -> str:
        """Convert object into a json string.

        Returns
        -------
        str
            The instance of the object in json string format
        """
        json_dict = self.to_json()
        return json.dumps(json_dict)



from typing import Dict, Any
from abc import ABC, abstractmethod


class Action(ABC):
    """Defines the base abstract class for Actions that need to be implemented."""

    @property
    @abstractmethod
    def action_type(self) -> int:
        """Get the type of action this class represents."""
        pass

    @abstractmethod
    def to_json(self) -> Dict[str, Any]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        pass



from furl import furl
from datetime import datetime, timedelta

from azure.durable_functions.models import OrchestrationRuntimeStatus
from azure.durable_functions.models.RpcManagementOptions import RpcManagementOptions
from azure.durable_functions.constants import DATETIME_STRING_FORMAT
from tests.test_utils.constants import RPC_BASE_URL


def assert_urls_match(expected, result):
    expected_url = furl(expected)
    result_url = furl(result)

    assert result_url.path == expected_url.path
    assert len(result_url.args) == len(expected_url.args)

    for arg in expected_url.args:
        assert expected_url.args[arg] == result_url.args[arg]


def test_just_instance_id():
    instance_id = 'test1234'
    options = RpcManagementOptions(instance_id=instance_id)
    result = options.to_url(RPC_BASE_URL)
    expected = f"{RPC_BASE_URL}instances/{instance_id}"
    assert_urls_match(expected=expected, result=result)


def test_instance_id_with_optional_booleans():
    instance_id = 'test1234'
    options = RpcManagementOptions(instance_id=instance_id, show_history=True,
                                   show_history_output=True, show_input=True)
    result = options.to_url(RPC_BASE_URL)
    expected = f"{RPC_BASE_URL}instances/{instance_id}?" \
               "showHistory=True&showHistoryOutput=True&showInput=True"

    assert_urls_match(expected=expected, result=result)


def test_just_the_strings():
    task_hub_name = 'my_hub'
    connection_name = 'my_connection'
    options = RpcManagementOptions(connection_name=connection_name, task_hub_name=task_hub_name)
    result = options.to_url(RPC_BASE_URL)
    expected = f"{RPC_BASE_URL}instances/?connectionName={connection_name}&taskHub={task_hub_name}"

    assert_urls_match(expected=expected, result=result)


def test_one_runtime_status():
    runtime_status = [OrchestrationRuntimeStatus.Running]
    options = RpcManagementOptions(runtime_status=runtime_status)
    result = options.to_url(RPC_BASE_URL)
    expected = f"{RPC_BASE_URL}instances/?runtimeStatus=Running"

    assert_urls_match(expected=expected, result=result)


def test_two_runtime_status():
    runtime_status = [OrchestrationRuntimeStatus.Running, OrchestrationRuntimeStatus.Completed]
    options = RpcManagementOptions(runtime_status=runtime_status)
    result = options.to_url(RPC_BASE_URL)
    expected = f"{RPC_BASE_URL}instances/?runtimeStatus=Running,Completed"

    assert_urls_match(expected=expected, result=result)


def test_datetime_status():
    created_time_from = datetime.now()
    created_time_to = created_time_from + timedelta(minutes=1)
    options = RpcManagementOptions(created_time_from=created_time_from,
                                   created_time_to=created_time_to)
    result = options.to_url(RPC_BASE_URL)
    from_as_string = created_time_from.strftime(DATETIME_STRING_FORMAT)
    to_as_string = created_time_to.strftime(DATETIME_STRING_FORMAT)
    expected = f"{RPC_BASE_URL}instances/?createdTimeFrom={from_as_string}" \
               f"&createdTimeTo={to_as_string}"

    assert_urls_match(expected=expected, result=result)



import uuid
from typing import Any, Dict

from tests.test_utils.json_utils import add_attrib


class OrchestrationInstance:
    def __init__(self):
        self.instance_id: str = str(uuid.uuid4())
        self.execution_id: str = str(uuid.uuid4())

    def to_json(self) -> Dict[str, Any]:
        json_dict = {}

        add_attrib(json_dict, self, 'instance_id', 'InstanceId')
        add_attrib(json_dict, self, 'execution_id', 'ExecutionId')

        return json_dict



import json
import logging
from typing import Any, Dict, Union, cast

import azure.functions as func
from azure.durable_functions import DurableOrchestrationClient
from azure.durable_functions.models.utils.entity_utils import EntityId


async def main(req: func.HttpRequest, starter: str) -> func.HttpResponse:
    client = DurableOrchestrationClient(starter)
    entity_name, entity_key = req.route_params["entityName"], req.route_params["entityKey"]
    
    entity_identifier = EntityId(entity_name, entity_key)

    entity_state_response = await client.read_entity_state(entity_identifier)

    if not entity_state_response.entity_exists:
        return func.HttpResponse("Entity not found", status_code=404)

    return func.HttpResponse(json.dumps({
        "entity": entity_name,
        "key": entity_key,
        "state": entity_state_response.entity_state
    }))
    


from collections import defaultdict
from azure.durable_functions.models.actions.SignalEntityAction import SignalEntityAction
from azure.durable_functions.models.actions.CallEntityAction import CallEntityAction
from azure.durable_functions.models.Task import TaskBase, TimerTask
from azure.durable_functions.models.actions.CallHttpAction import CallHttpAction
from azure.durable_functions.models.DurableHttpRequest import DurableHttpRequest
from azure.durable_functions.models.actions.CallSubOrchestratorWithRetryAction import \
    CallSubOrchestratorWithRetryAction
from azure.durable_functions.models.actions.CallActivityWithRetryAction import \
    CallActivityWithRetryAction
from azure.durable_functions.models.actions.ContinueAsNewAction import \
    ContinueAsNewAction
from azure.durable_functions.models.actions.WaitForExternalEventAction import \
    WaitForExternalEventAction
from azure.durable_functions.models.actions.CallSubOrchestratorAction import \
    CallSubOrchestratorAction
from azure.durable_functions.models.actions.CreateTimerAction import CreateTimerAction
from azure.durable_functions.models.Task import WhenAllTask, WhenAnyTask, AtomicTask, \
    RetryAbleTask
from azure.durable_functions.models.actions.CallActivityAction import CallActivityAction
from azure.durable_functions.models.ReplaySchema import ReplaySchema
import json
import datetime
import inspect
from typing import DefaultDict, List, Any, Dict, Optional, Tuple, Union, Callable
from uuid import UUID, uuid5, NAMESPACE_URL, NAMESPACE_OID
from datetime import timezone

from .RetryOptions import RetryOptions
from .FunctionContext import FunctionContext
from .history import HistoryEvent, HistoryEventType
from .actions import Action
from ..models.TokenSource import TokenSource
from .utils.entity_utils import EntityId
from azure.functions._durable_functions import _deserialize_custom_object
from azure.durable_functions.constants import DATETIME_STRING_FORMAT
from azure.durable_functions.decorators.metadata import OrchestrationTrigger, ActivityTrigger
from azure.functions.decorators.function_app import FunctionBuilder


class DurableOrchestrationContext:
    """Context of the durable orchestration execution.

    Parameter data for orchestration bindings that can be used to schedule
    function-based activities.
    """

    # parameter names are as defined by JSON schema and do not conform to PEP8 naming conventions
    def __init__(self,
                 history: List[Dict[Any, Any]], instanceId: str, isReplaying: bool,
                 parentInstanceId: str, input: Any = None, upperSchemaVersion: int = 0, **kwargs):
        self._histories: List[HistoryEvent] = [HistoryEvent(**he) for he in history]
        self._instance_id: str = instanceId
        self._is_replaying: bool = isReplaying
        self._parent_instance_id: str = parentInstanceId
        self._custom_status: Any = None
        self._new_uuid_counter: int = 0
        self._sub_orchestrator_counter: int = 0
        self._continue_as_new_flag: bool = False
        self.decision_started_event: HistoryEvent = \
            [e_ for e_ in self.histories
             if e_.event_type == HistoryEventType.ORCHESTRATOR_STARTED][0]
        self._current_utc_datetime: datetime.datetime = \
            self.decision_started_event.timestamp
        self._new_uuid_counter = 0
        self._function_context: FunctionContext = FunctionContext(**kwargs)
        self._sequence_number = 0
        self._replay_schema = ReplaySchema(upperSchemaVersion)

        self._action_payload_v1: List[List[Action]] = []
        self._action_payload_v2: List[Action] = []

        # make _input always a string
        # (consistent with Python Functions generic trigger/input bindings)
        if (isinstance(input, Dict)):
            input = json.dumps(input)

        self._input: Any = input
        self.open_tasks: DefaultDict[Union[int, str], Union[List[TaskBase], TaskBase]]
        self.open_tasks = defaultdict(list)
        self.deferred_tasks: Dict[Union[int, str], Tuple[HistoryEvent, bool, str]] = {}

    @classmethod
    def from_json(cls, json_string: str):
        """Convert the value passed into a new instance of the class.

        Parameters
        ----------
        json_string: str
            Context passed a JSON serializable value to be converted into an instance of the class

        Returns
        -------
        DurableOrchestrationContext
            New instance of the durable orchestration context class
        """
        # We should consider parsing the `Input` field here as well,
        # instead of doing so lazily when `get_input` is called.
        json_dict = json.loads(json_string)
        return cls(**json_dict)

    def _generate_task(self, action: Action,
                       retry_options: Optional[RetryOptions] = None,
                       id_: Optional[Union[int, str]] = None,
                       parent: Optional[TaskBase] = None,
                       task_constructor=AtomicTask) -> Union[AtomicTask, RetryAbleTask, TimerTask]:
        """Generate an atomic or retryable Task based on an input.

        Parameters
        ----------
        action : Action
            The action backing the Task.
        retry_options : Optional[RetryOptions]
            RetryOptions for a with-retry task, by default None

        Returns
        -------
        Union[AtomicTask, RetryAbleTask]
            Either an atomic task or a retry-able task
        """
        # Create an atomic task
        task: Union[AtomicTask, RetryAbleTask]
        action_payload: Union[Action, List[Action]]

        # TODO: find cleanear way to do this
        if self._replay_schema is ReplaySchema.V1:
            action_payload = [action]
        else:
            action_payload = action
        task = task_constructor(id_, action_payload)
        task.parent = parent

        # if task is retryable, provide the retryable wrapper class
        if not (retry_options is None):
            task = RetryAbleTask(task, retry_options, self)
        return task

    def _set_is_replaying(self, is_replaying: bool):
        """Set the internal `is_replaying` flag.

        Parameters
        ----------
        is_replaying : bool
            New value of the `is_replaying` flag
        """
        self._is_replaying = is_replaying

    def call_activity(self, name: Union[str, Callable], input_: Optional[Any] = None) -> TaskBase:
        """Schedule an activity for execution.

        Parameters
        ----------
        name: str | Callable
            Either the name of the activity function to call, as a string or,
            in the Python V2 programming model, the activity function itself.
        input_: Optional[Any]
            The JSON-serializable input to pass to the activity function.

        Returns
        -------
        Task
            A Durable Task that completes when the called activity function completes or fails.
        """
        if isinstance(name, Callable) and not isinstance(name, FunctionBuilder):
            error_message = "The `call_activity` API received a `Callable` without an "\
                "associated Azure Functions trigger-type. "\
                "Please ensure you're using the Python programming model V2 "\
                "and that your activity function is annotated with the `activity_trigger`"\
                "decorator. Otherwise, provide in the name of the activity as a string."
            raise ValueError(error_message)

        if isinstance(name, FunctionBuilder):
            name = self._get_function_name(name, ActivityTrigger)

        action = CallActivityAction(name, input_)
        task = self._generate_task(action)
        return task

    def call_activity_with_retry(self,
                                 name: Union[str, Callable], retry_options: RetryOptions,
                                 input_: Optional[Any] = None) -> TaskBase:
        """Schedule an activity for execution with retry options.

        Parameters
        ----------
        name: str | Callable
            Either the name of the activity function to call, as a string or,
            in the Python V2 programming model, the activity function itself.
        retry_options: RetryOptions
            The retry options for the activity function.
        input_: Optional[Any]
            The JSON-serializable input to pass to the activity function.

        Returns
        -------
        Task
            A Durable Task that completes when the called activity function completes or
            fails completely.
        """
        if isinstance(name, Callable) and not isinstance(name, FunctionBuilder):
            error_message = "The `call_activity` API received a `Callable` without an "\
                "associated Azure Functions trigger-type. "\
                "Please ensure you're using the Python programming model V2 "\
                "and that your activity function is annotated with the `activity_trigger`"\
                "decorator. Otherwise, provide in the name of the activity as a string."
            raise ValueError(error_message)

        if isinstance(name, FunctionBuilder):
            name = self._get_function_name(name, ActivityTrigger)

        action = CallActivityWithRetryAction(name, retry_options, input_)
        task = self._generate_task(action, retry_options)
        return task

    def call_http(self, method: str, uri: str, content: Optional[str] = None,
                  headers: Optional[Dict[str, str]] = None,
                  token_source: TokenSource = None) -> TaskBase:
        """Schedule a durable HTTP call to the specified endpoint.

        Parameters
        ----------
        method: str
            The HTTP request method.
        uri: str
            The HTTP request uri.
        content: Optional[str]
            The HTTP request content.
        headers: Optional[Dict[str, str]]
            The HTTP request headers.
        token_source: TokenSource
            The source of OAuth token to add to the request.

        Returns
        -------
        Task
            The durable HTTP request to schedule.
        """
        json_content: Optional[str] = None
        if content and content is not isinstance(content, str):
            json_content = json.dumps(content)
        else:
            json_content = content

        request = DurableHttpRequest(method, uri, json_content, headers, token_source)
        action = CallHttpAction(request)
        task = self._generate_task(action)
        return task

    def call_sub_orchestrator(self,
                              name: Union[str, Callable], input_: Optional[Any] = None,
                              instance_id: Optional[str] = None) -> TaskBase:
        """Schedule sub-orchestration function named `name` for execution.

        Parameters
        ----------
        name: Union[str, Callable]
            The name of the orchestrator function to call.
        input_: Optional[Any]
            The JSON-serializable input to pass to the orchestrator function.
        instance_id: Optional[str]
            A unique ID to use for the sub-orchestration instance.

        Returns
        -------
        Task
            A Durable Task that completes when the called sub-orchestrator completes or fails.
        """
        if isinstance(name, Callable) and not isinstance(name, FunctionBuilder):
            error_message = "The `call_activity` API received a `Callable` without an "\
                "associated Azure Functions trigger-type. "\
                "Please ensure you're using the Python programming model V2 "\
                "and that your activity function is annotated with the `activity_trigger`"\
                "decorator. Otherwise, provide in the name of the activity as a string."
            raise ValueError(error_message)

        if isinstance(name, FunctionBuilder):
            name = self._get_function_name(name, OrchestrationTrigger)

        action = CallSubOrchestratorAction(name, input_, instance_id)
        task = self._generate_task(action)
        return task

    def call_sub_orchestrator_with_retry(self,
                                         name: Union[str, Callable], retry_options: RetryOptions,
                                         input_: Optional[Any] = None,
                                         instance_id: Optional[str] = None) -> TaskBase:
        """Schedule sub-orchestration function named `name` for execution, with retry-options.

        Parameters
        ----------
        name: Union[str, Callable]
            The name of the activity function to schedule.
        retry_options: RetryOptions
            The settings for retrying this sub-orchestrator in case of a failure.
        input_: Optional[Any]
            The JSON-serializable input to pass to the activity function. Defaults to None.
        instance_id: str
            The instance ID of the sub-orchestrator to call.

        Returns
        -------
        Task
            A Durable Task that completes when the called sub-orchestrator completes or fails.
        """
        if isinstance(name, Callable) and not isinstance(name, FunctionBuilder):
            error_message = "The `call_activity` API received a `Callable` without an "\
                "associated Azure Functions trigger-type. "\
                "Please ensure you're using the Python programming model V2 "\
                "and that your activity function is annotated with the `activity_trigger`"\
                "decorator. Otherwise, provide in the name of the activity as a string."
            raise ValueError(error_message)

        if isinstance(name, FunctionBuilder):
            name = self._get_function_name(name, OrchestrationTrigger)

        action = CallSubOrchestratorWithRetryAction(name, retry_options, input_, instance_id)
        task = self._generate_task(action, retry_options)
        return task

    def get_input(self) -> Optional[Any]:
        """Get the orchestration input."""
        return None if self._input is None else json.loads(self._input,
                                                           object_hook=_deserialize_custom_object)

    def new_uuid(self) -> str:
        """Create a new UUID that is safe for replay within an orchestration or operation.

        The default implementation of this method creates a name-based UUID
        using the algorithm from RFC 4122 §4.3. The name input used to generate
        this value is a combination of the orchestration instance ID and an
        internally managed sequence number.

        Returns
        -------
        str
            New UUID that is safe for replay within an orchestration or operation.
        """
        URL_NAMESPACE: str = "9e952958-5e33-4daf-827f-2fa12937b875"

        uuid_name_value = \
            f"{self._instance_id}" \
            f"_{self.current_utc_datetime.strftime(DATETIME_STRING_FORMAT)}" \
            f"_{self._new_uuid_counter}"
        self._new_uuid_counter += 1
        namespace_uuid = uuid5(NAMESPACE_OID, URL_NAMESPACE)
        return str(uuid5(namespace_uuid, uuid_name_value))

    def task_all(self, activities: List[TaskBase]) -> TaskBase:
        """Schedule the execution of all activities.

        Similar to Promise.all. When called with `yield` or `return`, returns an
        array containing the results of all [[Task]]s passed to it. It returns
        when all of the [[Task]] instances have completed.

        Throws an exception if any of the activities fails
        Parameters
        ----------
        activities: List[Task]
            List of activities to schedule

        Returns
        -------
        TaskSet
            The results of all activities.
        """
        return WhenAllTask(activities, replay_schema=self._replay_schema)

    def task_any(self, activities: List[TaskBase]) -> TaskBase:
        """Schedule the execution of all activities.

        Similar to Promise.race. When called with `yield` or `return`, returns
        the first [[Task]] instance to complete.

        Throws an exception if all of the activities fail

        Parameters
        ----------
        activities: List[Task]
            List of activities to schedule

        Returns
        -------
        TaskSet
            The first [[Task]] instance to complete.
        """
        return WhenAnyTask(activities, replay_schema=self._replay_schema)

    def set_custom_status(self, status: Any):
        """Set the customized orchestration status for your orchestrator function.

        This status is also returned by the orchestration client through the get_status API

        Parameters
        ----------
        status : str
            Customized status provided by the orchestrator
        """
        self._custom_status = status

    @property
    def custom_status(self):
        """Get customized status of current orchestration."""
        return self._custom_status

    @property
    def histories(self):
        """Get running history of tasks that have been scheduled."""
        return self._histories

    @property
    def instance_id(self) -> str:
        """Get the ID of the current orchestration instance.

        The instance ID is generated and fixed when the orchestrator function
        is scheduled. It can be either auto-generated, in which case it is
        formatted as a GUID, or it can be user-specified with any format.

        Returns
        -------
        str
            The ID of the current orchestration instance.
        """
        return self._instance_id

    @property
    def is_replaying(self) -> bool:
        """Get the value indicating orchestration replaying itself.

        This property is useful when there is logic that needs to run only when
        the orchestrator function is _not_ replaying. For example, certain
        types of application logging may become too noisy when duplicated as
        part of orchestrator function replay. The orchestrator code could check
        to see whether the function is being replayed and then issue the log
        statements when this value is `false`.

        Returns
        -------
        bool
            Value indicating whether the orchestrator function is currently replaying.
        """
        return self._is_replaying

    @property
    def parent_instance_id(self) -> str:
        """Get the ID of the parent orchestration.

        The parent instance ID is generated and fixed when the parent
        orchestrator function is scheduled. It can be either auto-generated, in
        which case it is formatted as a GUID, or it can be user-specified with
        any format.

        Returns
        -------
        str
            ID of the parent orchestration of the current sub-orchestration instance
        """
        return self._parent_instance_id

    @property
    def current_utc_datetime(self) -> datetime.datetime:
        """Get the current date/time.

        This date/time value is derived from the orchestration history. It
        always returns the same value at specific points in the orchestrator
        function code, making it deterministic and safe for replay.

        Returns
        -------
        datetime
            The current date/time in a way that is safe for use by orchestrator functions
        """
        return self._current_utc_datetime

    @current_utc_datetime.setter
    def current_utc_datetime(self, value: datetime.datetime):
        self._current_utc_datetime = value

    @property
    def function_context(self) -> FunctionContext:
        """Get the function level attributes not used by durable orchestrator.

        Returns
        -------
        FunctionContext
            Object containing function level attributes not used by durable orchestrator.
        """
        return self._function_context

    def call_entity(self, entityId: EntityId,
                    operationName: str, operationInput: Optional[Any] = None):
        """Get the result of Durable Entity operation given some input.

        Parameters
        ----------
        entityId: EntityId
            The ID of the entity to call
        operationName: str
            The operation to execute
        operationInput: Optional[Any]
            The input for tne operation, defaults to None.

        Returns
        -------
        Task
            A Task of the entity call
        """
        action = CallEntityAction(entityId, operationName, operationInput)
        task = self._generate_task(action)
        return task

    def _record_fire_and_forget_action(self, action: Action):
        """Append a responseless-API action object to the actions array.

        Parameters
        ----------
        action : Action
            The action to append
        """
        new_action: Union[List[Action], Action]
        if self._replay_schema is ReplaySchema.V2:
            new_action = action
        else:
            new_action = [action]
        self._add_to_actions(new_action)
        self._sequence_number += 1

    def signal_entity(self, entityId: EntityId,
                      operationName: str, operationInput: Optional[Any] = None):
        """Send a signal operation to Durable Entity given some input.

        Parameters
        ----------
        entityId: EntityId
            The ID of the entity to call
        operationName: str
            The operation to execute
        operationInput: Optional[Any]
            The input for tne operation, defaults to None.

        Returns
        -------
        Task
            A Task of the entity signal
        """
        action = SignalEntityAction(entityId, operationName, operationInput)
        task = self._generate_task(action)
        self._record_fire_and_forget_action(action)
        return task

    @property
    def will_continue_as_new(self) -> bool:
        """Return true if continue_as_new was called."""
        return self._continue_as_new_flag

    def create_timer(self, fire_at: datetime.datetime) -> TaskBase:
        """Create a Timer Task to fire after at the specified deadline.

        Parameters
        ----------
        fire_at : datetime.datetime
            The time for the timer to trigger

        Returns
        -------
        TaskBase
            A Durable Timer Task that schedules the timer to wake up the activity
        """
        action = CreateTimerAction(fire_at)
        task = self._generate_task(action, task_constructor=TimerTask)
        return task

    def wait_for_external_event(self, name: str) -> TaskBase:
        """Wait asynchronously for an event to be raised with the name `name`.

        Parameters
        ----------
        name : str
            The event name of the event that the task is waiting for.

        Returns
        -------
        Task
            Task to wait for the event
        """
        action = WaitForExternalEventAction(name)
        task = self._generate_task(action, id_=name)
        return task

    def continue_as_new(self, input_: Any):
        """Schedule the orchestrator to continue as new.

        Parameters
        ----------
        input_ : Any
            The new starting input to the orchestrator.
        """
        continue_as_new_action: Action = ContinueAsNewAction(input_)
        self._record_fire_and_forget_action(continue_as_new_action)
        self._continue_as_new_flag = True

    def new_guid(self) -> UUID:
        """Generate a replay-safe GUID.

        Returns
        -------
        UUID
            A new globally-unique ID
        """
        guid_name = f"{self.instance_id}_{self.current_utc_datetime}"\
            f"_{self._new_uuid_counter}"
        self._new_uuid_counter += 1
        guid = uuid5(NAMESPACE_URL, guid_name)
        return guid

    @property
    def _actions(self) -> List[List[Action]]:
        """Get the actions payload of this context, for replay in the extension.

        Returns
        -------
        List[List[Action]]
            The actions of this context
        """
        if self._replay_schema is ReplaySchema.V1:
            return self._action_payload_v1
        else:
            return [self._action_payload_v2]

    def _add_to_actions(self, action_repr: Union[List[Action], Action]):
        """Add a Task's actions payload to the context's actions array.

        Parameters
        ----------
        action_repr : Union[List[Action], Action]
            The tasks to add
        """
        # Do not add further actions after `continue_as_new` has been
        # called
        if self.will_continue_as_new:
            return

        if self._replay_schema is ReplaySchema.V1 and isinstance(action_repr, list):
            self._action_payload_v1.append(action_repr)
        elif self._replay_schema is ReplaySchema.V2 and isinstance(action_repr, Action):
            self._action_payload_v2.append(action_repr)
        else:
            raise Exception(f"DF-internal exception: ActionRepr of signature {type(action_repr)}"
                            f"is not compatible on ReplaySchema {self._replay_schema.name}. ")

    def _pretty_print_history(self) -> str:
        """Get a pretty-printed version of the orchestration's internal history."""
        def history_to_string(event):
            json_dict = {}
            for key, val in inspect.getmembers(event):
                if not key.startswith('_') and not inspect.ismethod(val):
                    if isinstance(val, datetime.date):
                        val = val.replace(tzinfo=timezone.utc).timetuple()
                    json_dict[key] = val
            return json.dumps(json_dict)
        return str(list(map(history_to_string, self._histories)))

    def _add_to_open_tasks(self, task: TaskBase):

        if task._is_scheduled:
            return

        if isinstance(task, AtomicTask):
            if task.id is None:
                task.id = self._sequence_number
                self._sequence_number += 1
                self.open_tasks[task.id] = task
            elif task.id != -1:
                self.open_tasks[task.id].append(task)

            if task.id in self.deferred_tasks:
                task_update_action = self.deferred_tasks[task.id]
                task_update_action()
        else:
            for child in task.children:
                self._add_to_open_tasks(child)

    def _get_function_name(self, name: FunctionBuilder,
                           trigger_type: Union[OrchestrationTrigger, ActivityTrigger]):
        try:
            if (isinstance(name._function._trigger, trigger_type)):
                name = name._function._name
                return name
            else:
                if (trigger_type == OrchestrationTrigger):
                    trigger_type = "OrchestrationTrigger"
                else:
                    trigger_type = "ActivityTrigger"
                error_message = "Received function with Trigger-type `"\
                    + name._function._trigger.type\
                    + "` but expected `" + trigger_type + "`. Ensure your "\
                    "function is annotated with the `" + trigger_type +\
                    "` decorator or directly pass in the name of the "\
                    "function as a string."
                raise ValueError(error_message)
        except AttributeError as e:
            e.message = "Durable Functions SDK internal error: an "\
                "expected attribute is missing from the `FunctionBuilder` "\
                "object in the Python V2 programming model. Please report "\
                "this bug in the Durable Functions Python SDK repo: "\
                "https://github.com/Azure/azure-functions-durable-python.\n"\
                "Error trace: " + e.message
            raise e



import logging

def main(num: int) -> int:
    """Activity function performing a specific step in the chain
    
    Parameters
    ----------
    num : int
        number whose value to increase by one
    
    Returns
    -------
    int
        the input, plus one
    """

    logging.info(f"Activity Triggered: {num}")
    return num + 1

    



import logging
import azure.functions as func
import azure.durable_functions as df

myApp = df.DFApp(http_auth_level=func.AuthLevel.ANONYMOUS)

@myApp.route(route="orchestrators/{functionName}")
@myApp.durable_client_input(client_name="client")
async def http_start(req: func.HttpRequest, client):
    function_name = req.route_params.get('functionName')
    instance_id = await client.start_new(function_name)
    
    logging.info(f"Started orchestration with ID = '{instance_id}'.")
    return client.create_check_status_response(req, instance_id)

@myApp.orchestration_trigger(context_name="context")
def my_orchestrator(context: df.DurableOrchestrationContext):
    result1 = yield context.call_activity('say_hello', "Tokyo")
    result2 = yield context.call_activity('say_hello', "Seattle")
    result3 = yield context.call_activity('say_hello', "London")
    return [result1, result2, result3]

@myApp.activity_trigger(input_name="city")
def say_hello(city: str) -> str:
    return f"Hello {city}!"


import logging
import azure.functions as func
import azure.durable_functions as df

# To learn more about blueprints in the Python prog model V2,
# see: https://learn.microsoft.com/en-us/azure/azure-functions/functions-reference-python?tabs=asgi%2Capplication-level&pivots=python-mode-decorators#blueprints

# Note, the `func` namespace does not contain Durable Functions triggers and bindings, so to register blueprints of
# DF we need to use the `df` package's version of blueprints.
bp = df.Blueprint()

# We define a standard function-chaining DF pattern

@bp.route(route="startOrchestrator")
@bp.durable_client_input(client_name="client")
async def start_orchestrator(req: func.HttpRequest, client):
    instance_id = await client.start_new("my_orchestrator")
    
    logging.info(f"Started orchestration with ID = '{instance_id}'.")
    return client.create_check_status_response(req, instance_id)

@bp.orchestration_trigger(context_name="context")
def my_orchestrator(context: df.DurableOrchestrationContext):
    result1 = yield context.call_activity('say_hello', "Tokyo")
    result2 = yield context.call_activity('say_hello', "Seattle")
    result3 = yield context.call_activity('say_hello', "London")
    return [result1, result2, result3]

@bp.activity_trigger(input_name="city")
def say_hello(city: str) -> str:
    return f"Hello {city}!"


from typing import Dict, Union

from .Action import Action
from .ActionType import ActionType
from ..utils.json_utils import add_attrib
from json import dumps
from azure.functions._durable_functions import _serialize_custom_object


class CallActivityAction(Action):
    """Defines the structure of the Call Activity object.

    Provides the information needed by the durable extension to be able to schedule the activity.
    """

    def __init__(self, function_name: str, input_=None):
        self.function_name: str = function_name
        # It appears that `.input_` needs to be JSON-serializable at this point
        self.input_ = dumps(input_, default=_serialize_custom_object)

        if not self.function_name:
            raise ValueError("function_name cannot be empty")

    @property
    def action_type(self) -> int:
        """Get the type of action this class represents."""
        return ActionType.CALL_ACTIVITY

    def to_json(self) -> Dict[str, Union[str, int]]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Union[str, int]]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Union[str, int]] = {}
        add_attrib(json_dict, self, 'action_type', 'actionType')
        add_attrib(json_dict, self, 'function_name', 'functionName')
        add_attrib(json_dict, self, 'input_', 'input')
        return json_dict



from enum import Enum


class OrchestrationRuntimeStatus(Enum):
    """The status of an orchestration instance."""

    Running = 'Running'
    """The orchestration instance has started running."""

    Completed = 'Completed'
    """The orchestration instance has completed normally."""

    ContinuedAsNew = 'ContinuedAsNew'
    """The orchestration instance has restarted itself with a new history.

    This is a transient state.
    """

    Failed = 'Failed'
    """The orchestration instance failed with an error."""

    Canceled = 'Canceled'
    """The orchestration was canceled gracefully."""

    Terminated = 'Terminated'
    """The orchestration instance was stopped abruptly."""

    Pending = 'Pending'
    """The orchestration instance has been scheduled but has not yet started running."""



import datetime
from dateutil.parser import parse as dt_parse
from .HistoryEventType import HistoryEventType


class HistoryEvent:
    """Used to communicate state relevant information from the durable extension to the client."""

    # parameter names are as defined by JSON schema and do not conform to PEP8 naming conventions
    def __init__(self, EventType: HistoryEventType, EventId: int, IsPlayed: bool, Timestamp: str,
                 **kwargs):
        self._event_type: HistoryEventType = EventType
        self._event_id: int = EventId
        self._is_played: bool = IsPlayed
        self._timestamp: datetime.datetime = dt_parse(Timestamp)
        self._is_processed: bool = False

        self.Name = None
        self.InstanceId = None
        self.TaskScheduledId = None
        self.Reason = None
        self.Details = None
        self.Input = None
        if kwargs is not None:
            for key, value in kwargs.items():
                self.__setattr__(key, value)

    @property
    def event_type(self) -> HistoryEventType:
        """Get the history event type property.

        Returns
        -------
        HistoryEventType
            The type of history event
        """
        return self._event_type

    @property
    def event_id(self) -> int:
        """Get the event ID property.

        Returns
        -------
        int
            The value that represents the event sequence
        """
        return self._event_id

    @property
    def is_played(self) -> bool:
        """Get the is played property.

        Returns
        -------
        bool
            Value indicating whether the event has been played
        """
        return self._is_played

    @property
    def is_processed(self) -> bool:
        """Get the is process property.

        Returns
        -------
        bool
            Value indicating whether the orchestrator has processed the event
        """
        return self._is_processed

    @is_processed.setter
    def is_processed(self, value: bool):
        """Set the is processed property.

        Parameters
        ----------
        bool
            Value to set the property to
        """
        self._is_processed = value

    @property
    def timestamp(self) -> datetime.datetime:
        """Get the timestamp property.

        Returns
        -------
        datetime
            Value indicating the the time the event occurred
        """
        return self._timestamp



import requests
import json

def main(repoID: str) -> str:

    # We use the GitHub API to count the number of open issues in the repo provided
    # Note that the GitHub API only displays at most 30 issues per response, so
    # the maximum number this activity will return is 30. That's enough for demo'ing purposes.
    [user, repo] = repoID.split("/")
    url = f"https://api.github.com/repos/{user}/{repo}/issues?state=open"
    res = requests.get(url)
    if res.status_code != 200:
        error_message = f"Could not find repo {user} under {repo}! API endpoint hit was: {url}"
        raise Exception(error_message)
    issues = json.loads(res.text)
    too_many_issues: bool = len(issues) >= 3
    return too_many_issues


from azure.durable_functions.models.ReplaySchema import ReplaySchema
from .orchestrator_test_utils \
    import assert_orchestration_state_equals, get_orchestration_state_result, assert_valid_schema
from tests.test_utils.ContextBuilder import ContextBuilder
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.models.RetryOptions import RetryOptions
from azure.durable_functions.models.actions.CallSubOrchestratorWithRetryAction \
    import CallSubOrchestratorWithRetryAction

RETRY_OPTIONS = RetryOptions(5000, 3)

def generator_function(context):
    outputs = []
    retry_options = RETRY_OPTIONS

    task1 = yield context.call_sub_orchestrator_with_retry("HelloSubOrchestrator", retry_options, "Tokyo")
    task2 = yield context.call_sub_orchestrator_with_retry("HelloSubOrchestrator", retry_options, "Seattle")
    task3 = yield context.call_sub_orchestrator_with_retry("HelloSubOrchestrator", retry_options, "London")

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs

def base_expected_state(output=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema.value)


def add_hello_suborch_action(state: OrchestratorState, input_: str):
    retry_options = RETRY_OPTIONS
    action = CallSubOrchestratorWithRetryAction(function_name='HelloSubOrchestrator',
        retry_options=retry_options,_input=input_)
    state.actions.append([action])


def add_hello_suborch_completed_events(
        context_builder: ContextBuilder, id_: int, result: str):
    context_builder.add_sub_orchestrator_started_event(name="HelloSubOrchestrator", id_=id_,input_="")
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_sub_orchestrator_completed_event(result=result, id_=id_)

def add_hello_suborch_failed_events(
        context_builder: ContextBuilder, id_: int, reason: str, details: str):
    context_builder.add_sub_orchestrator_started_event(name="HelloSubOrchestrator", id_=id_,input_="")
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_sub_orchestrator_failed_event(id_=id_, reason=reason, details=details)

def add_retry_timer_events(context_builder: ContextBuilder, id_: int):
    fire_at = context_builder.add_timer_created_event(id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_timer_fired_event(id_=id_, fire_at=fire_at)

def test_tokyo_and_seattle_and_london_state():
    context_builder = ContextBuilder('test_simple_function')
    add_hello_suborch_completed_events(context_builder, 0, "\"Hello Tokyo!\"")
    add_hello_suborch_completed_events(context_builder, 1, "\"Hello Seattle!\"")
    add_hello_suborch_completed_events(context_builder, 2, "\"Hello London!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state(
        ['Hello Tokyo!', 'Hello Seattle!', 'Hello London!'])
    add_hello_suborch_action(expected_state, 'Tokyo')
    add_hello_suborch_action(expected_state, 'Seattle')
    add_hello_suborch_action(expected_state, 'London')
    expected_state._is_done = True
    expected = expected_state.to_json()

    #assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_tokyo_and_seattle_and_london_state_partial_failure():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_suborch_completed_events(context_builder, 0, "\"Hello Tokyo!\"")
    add_hello_suborch_failed_events(context_builder, 1, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 2)
    add_hello_suborch_completed_events(context_builder, 3, "\"Hello Seattle!\"")
    add_hello_suborch_completed_events(context_builder, 4, "\"Hello London!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state(
        ['Hello Tokyo!', 'Hello Seattle!', 'Hello London!'])
    add_hello_suborch_action(expected_state, 'Tokyo')
    add_hello_suborch_action(expected_state, 'Seattle')
    add_hello_suborch_action(expected_state, 'London')
    expected_state._is_done = True
    expected = expected_state.to_json()

    #assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)

def test_tokyo_and_seattle_and_london_state_all_failed():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_suborch_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_suborch_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)
    add_hello_suborch_failed_events(context_builder, 4, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 5)

    try:
        result = get_orchestration_state_result(
            context_builder, generator_function)
        # Should have error'ed out
        assert False
    except Exception as e:
        error_label = "\n\n$OutOfProcData$:"
        error_str = str(e)
    
        expected_state = base_expected_state()
        add_hello_suborch_action(expected_state, 'Tokyo')

        error_msg = f'{failed_reason} \n {failed_details}'
        expected_state._error = error_msg
        state_str = expected_state.to_json_string()
        
        expected_error_str = f"{error_msg}{error_label}{state_str}"
        assert expected_error_str == error_str


from typing import Dict, Any

from ...constants import DATETIME_STRING_FORMAT


def add_attrib(json_dict: Dict[str, Any], object_,
               attribute_name: str, alt_name: str = None):
    """Add the value of the attribute from the object to the dictionary.

    Used to dynamically add the value of the attribute if the value is present.

    Parameters
    ----------
    json_dict: The dictionary to add the attribute to
    object_: The object to look for the attribute on
    attribute_name: The name of the attribute to look for
    alt_name: An alternate name to provide to the attribute in the in the dictionary
    """
    if hasattr(object_, attribute_name):
        json_dict[alt_name or attribute_name] = \
            getattr(object_, attribute_name)


def add_datetime_attrib(json_dict: Dict[str, Any], object_,
                        attribute_name: str, alt_name: str = None):
    """Add the value of the attribute from the object to the dictionary converted into a string.

    Parameters
    ----------
    json_dict: The dictionary to add the attribute to
    object_: The object to look for the attribute on
    attribute_name: The name of the attribute to look for
    alt_name: An alternate name to provide to the attribute in the in the dictionary
    """
    if hasattr(object_, attribute_name):
        json_dict[alt_name or attribute_name] = \
            getattr(object_, attribute_name).strftime(DATETIME_STRING_FORMAT)


def add_json_attrib(json_dict: Dict[str, Any], object_,
                    attribute_name: str, alt_name: str = None):
    """Add the results of the to_json() function call of the attribute from the object to the dict.

    Used to dynamically add the JSON converted value of the attribute if the value is present.

    Parameters
    ----------
    json_dict: The dictionary to add the attribute to
    object_: The object to look for the attribute on
    attribute_name: The name of the attribute to look for
    alt_name: An alternate name to provide to the attribute in the in the dictionary
    """
    if hasattr(object_, attribute_name):
        attribute_value = getattr(object_, attribute_name)
        if attribute_value:
            json_dict[alt_name or attribute_name] = attribute_value.to_json()



import azure.functions as func
import azure.durable_functions as df


def orchestrator_function(context: df.DurableOrchestrationContext):
    result1 = yield context.call_activity('E1_SayHello', "Tokyo")
    result2 = yield context.call_activity('E1_SayHello', "Seattle")
    result3 = yield context.call_activity('E1_SayHello', "London")
    return [result1, result2, result3]

main = df.Orchestrator.create(orchestrator_function)


import pytest
import json
from dateutil.parser import parse as dt_parse
from azure.durable_functions.models.ReplaySchema import ReplaySchema

from azure.durable_functions.models.DurableOrchestrationContext \
    import DurableOrchestrationContext
from tests.test_utils.ContextBuilder import ContextBuilder


@pytest.fixture
def starting_context():
    context = DurableOrchestrationContext.from_json(
        '{"history":[{"EventType":12,"EventId":-1,"IsPlayed":false,'
        '"Timestamp":"2019-12-08T23:18:41.3240927Z"}, '
        '{"OrchestrationInstance":{'
        '"InstanceId":"48d0f95957504c2fa579e810a390b938", '
        '"ExecutionId":"fd183ee02e4b4fd18c95b773cfb5452b"},"EventType":0,'
        '"ParentInstance":null, '
        '"Name":"DurableOrchestratorTrigger","Version":"","Input":"null",'
        '"Tags":null,"EventId":-1,"IsPlayed":false, '
        '"Timestamp":"2019-12-08T23:18:39.756132Z"}],"input":null,'
        '"instanceId":"48d0f95957504c2fa579e810a390b938", '
        '"isReplaying":false,"parentInstanceId":null} ')
    return context

@pytest.fixture
def starting_context_v2():
    context = DurableOrchestrationContext.from_json(
        '{"history":[{"EventType":12,"EventId":-1,"IsPlayed":false,'
        '"Timestamp":"2019-12-08T23:18:41.3240927Z"}, '
        '{"OrchestrationInstance":{'
        '"InstanceId":"48d0f95957504c2fa579e810a390b938", '
        '"ExecutionId":"fd183ee02e4b4fd18c95b773cfb5452b"},"EventType":0,'
        '"ParentInstance":null, '
        '"Name":"DurableOrchestratorTrigger","Version":"","Input":"null",'
        '"Tags":null,"EventId":-1,"IsPlayed":false, '
        '"Timestamp":"2019-12-08T23:18:39.756132Z"}],"input":null,'
        '"instanceId":"48d0f95957504c2fa579e810a390b938", '
        '"upperSchemaVersion": 1, '
        '"isReplaying":false,"parentInstanceId":null} ')
    return context


def test_extracts_is_replaying(starting_context):
    assert not starting_context.is_replaying

def test_assumes_v1_replayschema(starting_context):
    assert starting_context._replay_schema is ReplaySchema.V1

def test_assumes_v2_replayschema(starting_context_v2):
    assert starting_context_v2._replay_schema is ReplaySchema.V2

def test_extracts_instance_id(starting_context):
    assert "48d0f95957504c2fa579e810a390b938" == starting_context.instance_id


def test_sets_current_utc_datetime(starting_context):
    assert \
        dt_parse("2019-12-08T23:18:41.3240927Z") == \
        starting_context.current_utc_datetime


def test_extracts_histories(starting_context):
    assert 2 == len(starting_context.histories)


def test_added_function_context_args():
    context_builder = ContextBuilder('test_function_context')

    additional_attributes = {"attrib1": 1, "attrib2": "two",
                             "attrib3": {"randomDictionary": "random"}}

    context_as_string = context_builder.to_json_string(**additional_attributes)

    durable_context = DurableOrchestrationContext.from_json(context_as_string)

    assert durable_context.function_context is not None
    for key in additional_attributes:
        assert additional_attributes[key] == getattr(durable_context.function_context, key)


def test_get_input_none(starting_context):
    test = starting_context.get_input()
    assert None == test


def test_get_input_string():
    builder = ContextBuilder('test_function_context')
    builder.input_ = json.dumps('Seattle')
    context = DurableOrchestrationContext.from_json(builder.to_json_string())
    assert 'Seattle' == context.get_input()


def test_get_input_json_str():
    builder = ContextBuilder('test_function_context')
    builder.input_ = json.dumps({ 'city': 'Seattle' })
    context = DurableOrchestrationContext.from_json(builder.to_json_string())

    result = context.get_input()

    assert 'Seattle' == result['city']



from datetime import datetime
from dateutil.parser import parse as dt_parse
from typing import Any, List, Dict, Optional, Union
from .OrchestrationRuntimeStatus import OrchestrationRuntimeStatus
from .utils.json_utils import add_attrib, add_datetime_attrib


class DurableOrchestrationStatus:
    """Represents the status of a durable orchestration instance.

    Can be fetched using [[DurableOrchestrationClient]].[[get_status]].
    """

    # parameter names are as defined by JSON schema and do not conform to PEP8 naming conventions
    def __init__(self, name: Optional[str] = None, instanceId: Optional[str] = None,
                 createdTime: Optional[str] = None, lastUpdatedTime: Optional[str] = None,
                 input: Optional[Any] = None, output: Optional[Any] = None,
                 runtimeStatus: Optional[OrchestrationRuntimeStatus] = None,
                 customStatus: Optional[Any] = None,
                 history: Optional[List[Any]] = None,
                 **kwargs):
        self._name: Optional[str] = name
        self._instance_id: Optional[str] = instanceId
        self._created_time: Optional[datetime] = \
            dt_parse(createdTime) if createdTime is not None else None
        self._last_updated_time: Optional[datetime] = dt_parse(lastUpdatedTime) \
            if lastUpdatedTime is not None else None
        self._input: Any = input
        self._output: Any = output
        self._runtime_status: Optional[OrchestrationRuntimeStatus] = runtimeStatus
        if runtimeStatus is not None:
            self._runtime_status = OrchestrationRuntimeStatus(runtimeStatus)
        self._custom_status: Any = customStatus
        self._history: Optional[List[Any]] = history
        if kwargs is not None:
            for key, value in kwargs.items():
                self.__setattr__(key, value)

    def __bool__(self):
        """Determine if a class resolves to True or False.

        We say that a DurableOrchestrationStatus if False if it has a value
        `None` for its `_created_time` value, which should be empty if it
        refers to a non-existent orchestration. This facilitates a clean
        implementation of the Singleton pattern

        Returns
        -------
        bool
            True if self._created_time is not None. False otherwise.
        """
        return self._created_time is not None

    @classmethod
    def from_json(cls, json_obj: Any):
        """Convert the value passed into a new instance of the class.

        Parameters
        ----------
        json_obj: any
            JSON object to be converted into an instance of the class

        Returns
        -------
        DurableOrchestrationStatus
            New instance of the durable orchestration status class
        """
        if isinstance(json_obj, str):
            return cls(message=json_obj)
        else:
            return cls(**json_obj)

    def to_json(self) -> Dict[str, Union[int, str]]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Union[int, str]]
            The instance of the class converted into a json dictionary
        """
        json: Dict[str, Union[int, str]] = {}
        add_attrib(json, self, 'name')
        add_attrib(json, self, 'instance_id', 'instanceId')
        add_datetime_attrib(json, self, 'created_time', 'createdTime')
        add_datetime_attrib(json, self, 'last_updated_time', 'lastUpdatedTime')
        add_attrib(json, self, 'output')
        add_attrib(json, self, 'input_', 'input')
        if self.runtime_status is not None:
            json["runtimeStatus"] = self.runtime_status.name
        add_attrib(json, self, 'custom_status', 'customStatus')
        add_attrib(json, self, 'history')
        return json

    @property
    def name(self) -> Optional[str]:
        """Get the orchestrator function name."""
        return self._name

    @property
    def instance_id(self) -> Optional[str]:
        """Get the unique ID of the instance.

        The instance ID is generated and fixed when the orchestrator
        function is scheduled. It can either auto-generated, in which case
        it is formatted as a UUID, or it can be user-specified with any
        format.
        """
        return self._instance_id

    @property
    def created_time(self) -> Optional[datetime]:
        """Get the time at which the orchestration instance was created.

        If the orchestration instance is in the [[Pending]] status, this
        time represents the time at which the orchestration instance was
        scheduled.
        """
        return self._created_time

    @property
    def last_updated_time(self) -> Optional[datetime]:
        """Get the time at which the orchestration instance last updated its execution history."""
        return self._last_updated_time

    @property
    def input_(self) -> Any:
        """Get the input of the orchestration instance."""
        return self._input

    @property
    def output(self) -> Any:
        """Get the output of the orchestration instance."""
        return self._output

    @property
    def runtime_status(self) -> Optional[OrchestrationRuntimeStatus]:
        """Get the runtime status of the orchestration instance."""
        return self._runtime_status

    @property
    def custom_status(self) -> Any:
        """Get the custom status payload (if any).

        Set by [[DurableOrchestrationContext]].[[set_custom_status]].
        """
        return self._custom_status

    @property
    def history(self) -> Optional[List[Any]]:
        """Get the execution history of the orchestration instance.

        The history log can be large and is therefore `undefined` by
        default. It is populated only when explicitly requested in the call
        to [[DurableOrchestrationClient]].[[get_status]].
        """
        return self._history



def assert_tasks_equal(task1, task2):
    assert task1.is_completed == task2.is_completed
    assert task1.is_faulted == task2.is_faulted
    assert task1.result == task2.result
    assert task1.timestamp == task2.timestamp
    assert task1.id == task2.id
    assert task1.action == task2.action
    assert str(task1.exception) == str(task2.exception)


def assert_taskset_equal(taskset1, taskset2):
    assert taskset1.is_completed == taskset2.is_completed
    assert taskset1.is_faulted == taskset2.is_faulted
    assert taskset1.result == taskset2.result
    assert taskset1.actions == taskset2.actions
    assert taskset1.timestamp == taskset2.timestamp
    assert str(taskset1.exception) == str(taskset2.exception)



from azure.durable_functions.models.TokenSource import ManagedIdentityTokenSource

def test_serialization_fields():
    """Validates the TokenSource contains the expected fields when serialized to JSON"""
    token_source = ManagedIdentityTokenSource(resource="TOKEN_SOURCE")
    token_source_json = token_source.to_json()

    # Output JSON should contain a resource field and a kind field set to `AzureManagedIdentity`
    assert "resource" in token_source_json.keys()
    assert "kind" in token_source_json.keys()
    assert token_source_json["kind"] == "AzureManagedIdentity"


import logging

import azure.functions as func
import azure.durable_functions as df


async def main(req: func.HttpRequest, starter: str) -> func.HttpResponse:
    client = df.DurableOrchestrationClient(starter)
    instance_id = await client.start_new(req.route_params["functionName"], None, None)

    logging.info(f"Started orchestration with ID = '{instance_id}'.")

    return client.create_check_status_response(req, instance_id)


from abc import ABC
from typing import Dict, Union

from azure.durable_functions.models.utils.json_utils import add_attrib


class TokenSource(ABC):
    """Token Source implementation for [Azure Managed Identities].

    https://docs.microsoft.com/azure/active-directory/managed-identities-azure-resources/overview.

    @example Get a list of Azure Subscriptions by calling the Azure Resource Manager HTTP API.
     ```python
    import azure.durable_functions as df

    def generator_function(context):
        return yield context.callHttp(
            "GET",
            "https://management.azure.com/subscriptions?api-version=2019-06-01",
            None,
            None,
            df.ManagedIdentityTokenSource("https://management.core.windows.net"))
    """

    def __init__(self):
        super().__init__()


class ManagedIdentityTokenSource(TokenSource):
    """Returns a `ManagedIdentityTokenSource` object."""

    def __init__(self, resource: str):
        super().__init__()
        self._resource: str = resource
        self._kind: str = "AzureManagedIdentity"

    @property
    def resource(self) -> str:
        """Get the Azure Active Directory resource identifier of the web API being invoked.

        For example, `https://management.core.windows.net/` or `https://graph.microsoft.com/`.
        """
        return self._resource

    def to_json(self) -> Dict[str, Union[str, int]]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Union[str, int]] = {}
        add_attrib(json_dict, self, 'resource')
        json_dict["kind"] = self._kind
        return json_dict



from typing import Dict, Union

from .Action import Action
from ..utils.json_utils import add_attrib
from typing import List
from abc import abstractmethod


class CompoundAction(Action):
    """Defines the structure of the WhenAll Action object.

    Provides the information needed by the durable extension to be able to invoke WhenAll tasks.
    """

    def __init__(self, compound_tasks: List[Action]):
        self.compound_tasks = compound_tasks

    @property
    @abstractmethod
    def action_type(self) -> int:
        """Get this object's action type as an integer."""
        ...

    def to_json(self) -> Dict[str, Union[str, int]]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Union[str, int]]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Union[str, int]] = {}
        add_attrib(json_dict, self, 'action_type', 'actionType')
        json_dict['compoundActions'] = list(map(lambda x: x.to_json(), self.compound_tasks))
        return json_dict



from typing import Any


class EntityStateResponse:
    """Entity state response object for [read_entity_state]."""

    def __init__(self, entity_exists: bool, entity_state: Any = None) -> None:
        self._entity_exists = entity_exists
        self._entity_state = entity_state

    @property
    def entity_exists(self) -> bool:
        """Get the bool representing whether entity exists."""
        return self._entity_exists

    @property
    def entity_state(self) -> Any:
        """Get the state of the entity.

        When [entity_exists] is False, this value will be None.
        Optional.
        """
        return self._entity_state



import logging

import azure.functions as func
import azure.durable_functions as df


def orchestrator_function(context: df.DurableOrchestrationContext):
    """This function provides the core function chaining orchestration logic
    and also sets custom status using which a user can see intermittent status of
    the orchestration through get_status client URL.

    Parameters
    ----------
    context: DurableOrchestrationContext
        This context has the past history
        and the durable orchestration API's to chain a set of functions

    Returns
    -------
    final_result: str
        Returns the final result after the chain completes

    Yields
    -------
    call_activity: str
        Yields at every step of the function chain orchestration logic
    """

    # Chained functions - output of a function is passed as
    # input to the next function in the chain
    r1 = yield context.call_activity("DurableActivity", 0)
    context.set_custom_status(f'{r1} ->')
    r2 = yield context.call_activity("DurableActivity", r1)
    context.set_custom_status(f'{r1} -> {r2} ->')
    r3 = yield context.call_activity("DurableActivity", r2)
    context.set_custom_status(f'{r1} -> {r2} -> {r3}')
    return r3

main = df.Orchestrator.create(orchestrator_function)



"""Utility functions used by the Durable Function python library.

_Internal Only_
"""
from pkgutil import extend_path
import typing
__path__: typing.Iterable[str] = extend_path(__path__, __name__)  # type: ignore
